# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

import pypowervm.exceptions as ex
import pypowervm.tests.test_fixtures as fx
from pypowervm.tests.test_utils import pvmhttp
import pypowervm.wrappers.job as jwrap

JOB_REQUEST_FILE = "job_request_power_off.txt"
JOB_RESPONSE_OK = "job_response_completed_ok.txt"
JOB_RESPONSE_FAILED = "job_response_completed_failed.txt"
JOB_RESPONSE_EXCEPTION = "job_response_exception.txt"

EXPECTED_ID = '1375391227297'
EXPECTED_STATUS = jwrap.JobStatus.COMPLETED_WITH_ERROR
EXPECTED_EXCEPTION_MESSAGE = 'This is an exception message'
EXPECTED_RESULTS_VALUE = 'This is an error message'
EXPECTED_GROUP_NAME = 'LogicalPartition'
EXPECTED_OPERATION_NAME = 'PowerOff'


class TestJobEntryWrapper(testtools.TestCase):

    def setUp(self):
        super(TestJobEntryWrapper, self).setUp()

        self.adpt = self.useFixture(fx.AdapterFx()).adpt

        def load(fname):
            resp = pvmhttp.load_pvm_resp(fname, adapter=self.adpt).response
            self.assertIsNotNone(resp, "Could not load %s " % fname)
            return jwrap.Job.wrap(resp)

        self._request_wrapper = load(JOB_REQUEST_FILE)
        self._ok_wrapper = load(JOB_RESPONSE_OK)
        self._failed_wrapper = load(JOB_RESPONSE_FAILED)
        self._exception_wrapper = load(JOB_RESPONSE_EXCEPTION)
        self._exception_wrapper.op = 'CLIRunner'

        # Create a bad wrapper to use when retrieving properties which don't
        # exist
        self._bad_wrapper = load(JOB_REQUEST_FILE)

        self.set_test_property_values()

    def set_single_value(self, entry, property_name, value):
        prop = entry.element.find(property_name)
        self.assertNotEqual(prop, None,
                            "Could not find property %s." % property_name)

        prop.text = str(value)

    def set_test_property_values(self):
        """Set expected values in entry so test code can work consistently."""
        self.set_single_value(self._ok_wrapper.entry,
                              jwrap._JOB_ID, EXPECTED_ID)
        self.set_single_value(self._request_wrapper.entry,
                              jwrap._JOB_GROUP_NAME,
                              EXPECTED_GROUP_NAME)
        self.set_single_value(self._request_wrapper.entry,
                              jwrap._JOB_OPERATION_NAME,
                              EXPECTED_OPERATION_NAME)
        self.set_single_value(self._failed_wrapper.entry,
                              jwrap._JOB_STATUS,
                              EXPECTED_STATUS)
        self.set_single_value(self._exception_wrapper.entry,
                              jwrap._JOB_MESSAGE,
                              EXPECTED_EXCEPTION_MESSAGE)
        # results value containing the message is the second one in a list
        props = self._failed_wrapper.entry.element.findall(
            jwrap._JOB_RESULTS_VALUE)
        props[1].text = str(EXPECTED_RESULTS_VALUE)

    def verify_equal(self, method_name, returned_value, expected_value):
        if returned_value is not None and expected_value is not None:
            returned_type = type(returned_value)
            expected_type = type(expected_value)
            self.assertEqual(returned_type, expected_type,
                             "%s: type mismatch.  "
                             "Returned %s(%s). Expected %s(%s)" %
                             (method_name,
                              returned_value, returned_type,
                              expected_value, expected_type))

        self.assertEqual(returned_value, expected_value,
                         "%s returned %s instead of %s"
                         % (method_name, returned_value, expected_value))

    def call_simple_getter(self,
                           wrapper,
                           method_name,
                           expected_value,
                           expected_bad_value):

        # Use __getattribute__ to dynamically call the method
        value = wrapper.__getattribute__(
            method_name)
        if callable(value):
            value = value()
        self.verify_equal(method_name, value, expected_value)

        bad_value = self._bad_wrapper.__getattribute__(method_name)
        if callable(bad_value):
            bad_value = bad_value()
        self.verify_equal(method_name, bad_value, expected_bad_value)

    def test_get_job_id(self):
        self.call_simple_getter(self._ok_wrapper,
                                "job_id", EXPECTED_ID, None)

    def test_get_job_status(self):
        self.call_simple_getter(self._failed_wrapper,
                                "job_status", EXPECTED_STATUS, None)

    def test_get_job_message(self):
        self.call_simple_getter(self._exception_wrapper,
                                "get_job_message",
                                EXPECTED_EXCEPTION_MESSAGE, '')

    def test_get_job_resp_exception_msg(self):
        self.call_simple_getter(self._exception_wrapper,
                                "get_job_resp_exception_msg",
                                EXPECTED_EXCEPTION_MESSAGE, '')

    def test_get_job_results_message(self):
        self.call_simple_getter(self._failed_wrapper,
                                "get_job_results_message",
                                EXPECTED_RESULTS_VALUE, '')

    def test_job_parameters(self):
        input_name1 = 'JobParmName1'
        input_name2 = 'JobParmName2'
        input_value1 = 'JobParmValue1'
        input_value2 = 'JobParmValue2'
        wrapper = self._request_wrapper
        job_parms = [
            wrapper.create_job_parameter(input_name1, input_value1),
            wrapper.create_job_parameter(input_name2, input_value2)]
        wrapper.add_job_parameters_to_existing(*job_parms)
        elements = wrapper.entry.element.findall(
            'JobParameters/JobParameter/ParameterName')
        names = []
        for element in elements:
            names.append(element.text)
        self.assertEqual(input_name1, names[0],
                         "Job names don't match")
        self.assertEqual(input_name2, names[1],
                         "Job names don't match")
        elements = wrapper.entry.element.findall(
            'JobParameters/JobParameter/ParameterValue')
        values = []
        for element in elements:
            values.append(element.text)
        self.assertEqual(input_value1, values[0],
                         "Job values don't match")
        self.assertEqual(input_value2, values[1],
                         "Job values don't match")

    @mock.patch('pypowervm.wrappers.job.Job._monitor_job')
    @mock.patch('pypowervm.wrappers.job.Job.cancel_job')
    @mock.patch('pypowervm.wrappers.job.Job.delete_job')
    @mock.patch('pypowervm.wrappers.job.Job.job_status')
    def test_run_job(self, mock_status, mock_del, mock_cancel, mock_monitor):
        mock_status.__get__ = mock.Mock(
            return_value=jwrap.JobStatus.COMPLETED_OK)
        wrapper = self._request_wrapper
        # Synchronous
        # No timeout
        mock_monitor.return_value = False
        wrapper.run_job('uuid')
        self.adpt.create_job.assert_called_with(mock.ANY, 'LogicalPartition',
                                                'uuid', sensitive=False)
        self.assertEqual(1, mock_monitor.call_count)
        self.assertEqual(0, mock_cancel.call_count)
        self.assertEqual(1, mock_del.call_count)

        # Time out
        mock_monitor.reset_mock()
        mock_del.reset_mock()
        mock_monitor.return_value = True
        self.assertRaises(ex.JobRequestTimedOut, wrapper.run_job, 'uuid')
        self.assertEqual(1, mock_monitor.call_count)
        self.assertEqual(1, mock_cancel.call_count)
        self.assertEqual(0, mock_del.call_count)

        # Non-OK status
        mock_status.__get__.return_value = jwrap.JobStatus.COMPLETED_WITH_ERROR
        mock_monitor.reset_mock()
        mock_cancel.reset_mock()
        mock_monitor.return_value = False
        self.assertRaises(ex.JobRequestFailed, wrapper.run_job, 'uuid')
        self.assertEqual(0, mock_cancel.call_count)
        self.assertEqual(1, mock_del.call_count)

        # Asynchronous.  With no timeout, return right after monitor.  "Bad"
        # result isn't checked, delete is not called.
        mock_del.reset_mock()
        wrapper.run_job('uuid', synchronous=False)
        self.assertEqual(0, mock_del.call_count)

    @mock.patch('pypowervm.wrappers.job.Job.poll_while_status')
    @mock.patch('pypowervm.wrappers.job.PollAndDeleteThread')
    def test_montor_job(self, mock_thread, mock_poll):
        wrapper = self._ok_wrapper
        # Synchronous is a pass-through to poll_while_status
        mock_poll.return_value = 'abc123'
        self.assertEqual('abc123', wrapper._monitor_job())
        mock_poll.assert_called_once_with(['RUNNING', 'NOT_STARTED'], mock.ANY,
                                          mock.ANY)
        self.assertEqual(0, mock_thread.call_count)

        # Asynchronous
        # Time out
        mock_poll.reset_mock()
        mock_poll.return_value = True
        self.assertTrue(wrapper._monitor_job(synchronous=False))
        mock_poll.assert_called_once_with(['NOT_STARTED'], mock.ANY, mock.ANY)
        self.assertEqual(0, mock_thread.call_count)

        # No timeout
        mock_poll.reset_mock()
        mock_poll.return_value = False
        thread_inst = mock.Mock()
        mock_thread.return_value = thread_inst
        self.assertFalse(wrapper._monitor_job(synchronous=False))
        mock_thread.assert_called_once_with(wrapper, False)
        thread_inst.start.assert_called_once_with()

    @mock.patch('time.time')
    @mock.patch('time.sleep')
    @mock.patch('pypowervm.wrappers.job.Job.job_status')
    def test_poll_while_status(self, mock_status, mock_sleep, mock_time):
        wrapper = self._ok_wrapper
        mock_status.__get__ = mock.Mock(return_value=jwrap.JobStatus.RUNNING)
        # Short-circuit if the status is already not in the list
        self.assertFalse(wrapper.poll_while_status(
            [jwrap.JobStatus.NOT_ACTIVE], 10, False))
        self.assertEqual(0, mock_sleep.call_count)
        self.assertEqual(1, mock_time.call_count)

        # Engineer a timeout after the third run
        mock_time.reset_mock()
        mock_time.side_effect = [0, 1, 2, 3, 4, 5]
        self.assertTrue(wrapper.poll_while_status(
            [jwrap.JobStatus.RUNNING], 3, False))
        self.assertEqual(3, mock_sleep.call_count)
        # Initial setup, bail on the fourth iteration
        self.assertEqual(5, mock_time.call_count)

        # "Infinite" timeout, status eventually becomes one not in the list
        mock_status.__get__.side_effect = ['a', 'b', 'c', 'd', 'e', 'f']
        mock_time.reset_mock()
        mock_time.side_effect = [0, 1, 2, 3, 4, 5]
        mock_sleep.reset_mock()
        self.assertFalse(wrapper.poll_while_status(['a', 'b', 'c', 'd', 'e'],
                                                   0, False))
        self.assertEqual(5, mock_sleep.call_count)
        # Only the initial timer setup
        self.assertEqual(6, mock_time.call_count)

    @mock.patch('pypowervm.wrappers.job.Job.poll_while_status')
    @mock.patch('pypowervm.wrappers.job.Job.delete_job')
    def test_poll_and_delete_thread(self, mock_del, mock_poll):
        # OK
        jwrap.PollAndDeleteThread(self._ok_wrapper, 'sens').run()
        mock_poll.assert_called_once_with(['RUNNING'], 0, 'sens')
        mock_del.assert_called_once_with()

        # Not OK
        mock_poll.reset_mock()
        mock_del.reset_mock()
        with self.assertLogs(jwrap.__name__, 'ERROR'):
            jwrap.PollAndDeleteThread(self._failed_wrapper, 'sens').run()
        mock_poll.assert_called_once_with(['RUNNING'], 0, 'sens')
        mock_del.assert_called_once_with()

    @mock.patch('pypowervm.wrappers.job.Job.poll_while_status')
    @mock.patch('pypowervm.wrappers.job.Job.delete_job')
    def test_cancel_job_thread(self, mock_del, mock_poll):
        jwrap.CancelJobThread(self._ok_wrapper, 'sens').run()
        mock_poll.assert_called_once_with(['RUNNING', 'NOT_STARTED'],
                                          0, 'sens')
        mock_del.assert_called_once_with()

    @mock.patch('pypowervm.wrappers.job.CancelJobThread.start')
    @mock.patch('pypowervm.wrappers.job.Job.delete_job')
    @mock.patch('pypowervm.wrappers.job.Job._monitor_job')
    def test_cancel_job(self, mock_monitor, mock_delete, mock_start):
        wrapper = self._ok_wrapper
        self.adpt.update.side_effect = ex.Error('error')
        mock_monitor.return_value = False
        wrapper.cancel_job()
        self.adpt.update.assert_called_with(
            None, None, root_type='jobs', root_id=wrapper.job_id,
            suffix_type='cancel')
        self.assertEqual(0, mock_delete.call_count)
        self.adpt.update.reset_mock()
        self.adpt.update.side_effect = None
        mock_start.reset_mock()
        mock_monitor.return_value = False
        wrapper.cancel_job()
        self.adpt.update.assert_called_with(
            None, None, root_type='jobs', root_id=wrapper.job_id,
            suffix_type='cancel')
        self.assertEqual(1, mock_start.call_count)

    @mock.patch('pypowervm.wrappers.job.Job.job_status')
    def test_delete_job(self, mock_status):
        wrapper = self._ok_wrapper
        mock_status.__get__ = mock.Mock(return_value=jwrap.JobStatus.RUNNING)
        with self.assertLogs(jwrap.__name__, 'ERROR'):
            self.assertRaises(ex.Error, wrapper.delete_job)
        mock_status.__get__.return_value = jwrap.JobStatus.COMPLETED_OK
        wrapper.delete_job()
        self.adpt.delete.assert_called_with('jobs', wrapper.job_id)
        self.adpt.delete.reset_mock()
        self.adpt.delete.side_effect = ex.Error('foo')
        with self.assertLogs(jwrap.__name__, 'ERROR'):
            wrapper.delete_job()
        self.adpt.delete.assert_called_with('jobs', wrapper.job_id)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Manage mappings of virtual storage devices from VIOS to LPAR."""

from oslo_concurrency import lockutils as lock
from oslo_log import log as logging

from pypowervm import const as c
from pypowervm.i18n import _
from pypowervm import util
from pypowervm.utils import retry as pvm_retry
from pypowervm.utils import uuid
from pypowervm.wrappers import storage as pvm_stor
from pypowervm.wrappers import virtual_io_server as pvm_vios

LOG = logging.getLogger(__name__)


def _argmod(this_try, max_tries, *args, **kwargs):
    """Retry argmod to change 'vios' arg from VIOS wrapper to a string UUID.

    This is so that etag mismatches trigger a fresh GET.
    """
    LOG.warning(_('Retrying modification of SCSI Mapping.'))
    argl = list(args)
    # Second argument is vios.
    if isinstance(argl[1], pvm_vios.VIOS):
        argl[1] = argl[1].uuid
    return argl, kwargs


@lock.synchronized('vscsi_mapping')
@pvm_retry.retry(tries=60, argmod_func=_argmod,
                 delay_func=pvm_retry.STEPPED_RANDOM_DELAY)
def add_vscsi_mapping(host_uuid, vios, lpar_uuid, storage_elem, fuse_limit=32,
                      lpar_slot_num=None, lua=None):
    """Will add a vSCSI mapping to a Virtual I/O Server.

    This method is used to connect a storage element (either a vDisk, vOpt,
    PV or LU) that resides on a Virtual I/O Server to a Virtual Machine.

    This is achieved using a 'vSCSI Mapping'.  The invoker does not need to
    interact with the mapping.

    A given mapping is essentially a 'vSCSI bus', which can host multiple
    storage elements.  This method has a fuse limit which throttles the number
    of devices on a given vSCSI bus.  The throttle should be lower if the
    storage elements are high I/O, and higher otherwise.

    :param host_uuid: The UUID of the host system.
    :param vios: The virtual I/O server to which the mapping should be
                 added.  This may be the VIOS's UUID string OR an existing VIOS
                 EntryWrapper.  If the latter, it must have been retrieved
                 using the VIO_SMAP extended attribute group.
    :param lpar_uuid: The UUID of the LPAR that will have the connected
                      storage.
    :param storage_elem: The storage element (either a vDisk, vOpt, LU or PV)
                         that is to be connected.
    :param fuse_limit: (Optional, Default: 32) The max number of devices to
                       allow on one scsi bus before creating a second SCSI bus.
    :param lpar_slot_num: (Optional, Default: None) The slot number for the
                          client LPAR to use in the mapping. If None, the next
                          available slot number is assigned by the server.
    :param lua: (Optional.  Default: None) Logical Unit Address to set on the
                TargetDevice.  If None, the LUA will be assigned by the server.
                Should be specified for all of the VSCSIMappings for a
                particular bus, or none of them.
    :return: The VIOS wrapper representing the updated Virtual I/O Server.
             This is current with respect to etag and SCSI mappings.
    """
    adapter = storage_elem.adapter

    # If the 'vios' param is a string UUID, retrieve the VIOS wrapper.
    if not isinstance(vios, pvm_vios.VIOS):
        vios_w = pvm_vios.VIOS.wrap(
            adapter.read(pvm_vios.VIOS.schema_type, root_id=vios,
                         xag=[c.XAG.VIO_SMAP]))
    else:
        vios_w = vios

    # If the storage element is already there, do nothing.
    if find_maps(vios_w.scsi_mappings, client_lpar_id=lpar_uuid,
                 stg_elem=storage_elem):
        LOG.info(_("Found existing mapping of %(stg_type)s storage element "
                   "%(stg_name)s from Virtual I/O Server %(vios_name)s to "
                   "client LPAR %(lpar_uuid)s."),
                 {'stg_type': storage_elem.schema_type,
                  'stg_name': storage_elem.name,
                  'vios_name': vios_w.name,
                  'lpar_uuid': lpar_uuid})
        return vios_w

    # Build the mapping.
    scsi_map = build_vscsi_mapping(host_uuid, vios_w, lpar_uuid, storage_elem,
                                   fuse_limit=fuse_limit,
                                   lpar_slot_num=lpar_slot_num, lua=lua)

    # Add the mapping.  It may have been updated to have a different client
    # and server adapter.  It may be the original (which creates a new client
    # and server pair).
    vios_w.scsi_mappings.append(scsi_map)

    LOG.info(_("Creating mapping of %(stg_type)s storage element %(stg_name)s "
               "from Virtual I/O Server %(vios_name)s to client LPAR "
               "%(lpar_uuid)s."),
             {'stg_type': storage_elem.schema_type,
              'stg_name': storage_elem.name,
              'vios_name': vios_w.name,
              'lpar_uuid': lpar_uuid})
    return vios_w.update()


def build_vscsi_mapping(host_uuid, vios_w, lpar_uuid, storage_elem,
                        fuse_limit=32, lpar_slot_num=None, lua=None,
                        target_name=None):
    """Will build a vSCSI mapping that can be added to a VIOS.

    This method is used to create a mapping element (for either a vDisk, vOpt,
    PV or LU) that connects a Virtual I/O Server to a LPAR.

    A given mapping is essentially a 'vSCSI bus', which can host multiple
    storage elements.  This method has a fuse limit which throttles the number
    of devices on a given vSCSI bus.  The throttle should be lower if the
    storage elements are high I/O, and higher otherwise.

    :param host_uuid: The UUID of the host system.
    :param vios_w: The virtual I/O server wrapper that the mapping is intended
                   to be attached to.  The method will call the update against
                   the API.  It will only update the in memory wrapper.
    :param lpar_uuid: The UUID of the LPAR that will have the connected
                      storage.
    :param storage_elem: The storage element (either a vDisk, vOpt, LU or PV)
                         that is to be connected.
    :param fuse_limit: (Optional, Default: 32) The max number of devices to
                       allow on one scsi bus before creating a second SCSI bus.
    :param lpar_slot_num: (Optional, Default: None) The slot number for the
                          client LPAR to use in the mapping. If None, the next
                          available slot number is assigned by the server.
    :param lua: (Optional.  Default: None) Logical Unit Address to set on the
                TargetDevice.  If None, the LUA will be assigned by the server.
                Should be specified for all of the VSCSIMappings for a
                particular bus, or none of them.
    :param target_name: (Optional, Default: None) The name of the Target
                        mapping. If None, the target_name will be assigned by
                        the server.
    :return: The SCSI mapping that can be added to the vios_w.  This does not
             do any updates to the wrapper itself.
    """
    adapter = storage_elem.adapter

    # Get the client lpar href
    lpar_href = pvm_vios.VSCSIMapping.crt_related_href(adapter, host_uuid,
                                                       lpar_uuid)

    # Separate out the mappings into the applicable ones for this client.
    separated_mappings = _separate_mappings(vios_w, lpar_href)

    # Used if we need to clone an existing mapping
    clonable_map = None

    # What we need to figure out is, within the existing mappings, can we
    # reuse the existing client and server adapter (which we can only do if
    # below the fuse limit), or if we need to create a new adapter pair.
    for mapping_list in separated_mappings.values():
        if len(mapping_list) < fuse_limit:
            # Swap in the first maps client/server adapters into the existing
            # map.  We call the semi-private methods as this is not something
            # that an 'update' would do...this is part of the 'create' flow.
            clonable_map = mapping_list[0]
            break

    # If we have a clonable map, we can replicate that.  Otherwise we need
    # to build from scratch.
    if clonable_map is not None:
        scsi_map = pvm_vios.VSCSIMapping.bld_from_existing(
            clonable_map, storage_elem, lpar_slot_num=lpar_slot_num, lua=lua,
            target_name=target_name)
    else:
        scsi_map = pvm_vios.VSCSIMapping.bld(
            adapter, host_uuid, lpar_uuid, storage_elem,
            lpar_slot_num=lpar_slot_num, lua=lua, target_name=target_name)
    return scsi_map


def _separate_mappings(vios_w, client_href):
    """Separates out the systems existing mappings into silos.

    :param vios_w: The pypowervm wrapper for the VIOS.
    :param client_href: The client to separate the mappings for.
    :return: A dictionary where the key is the server adapter (which is
             bound to the client).  The value is the list mappings that use
             the server adapter.
    """
    # The key is server_adapter.udid, the value is the list of applicable
    # mappings to the server adapter.
    resp = {}

    existing_mappings = vios_w.scsi_mappings
    for existing_map in existing_mappings:
        if (existing_map.client_lpar_href == client_href and
                # ignore orphaned mappings
                existing_map.client_adapter is not None):
            # Valid map to consider
            key = existing_map.server_adapter.udid
            if resp.get(key) is None:
                resp[key] = []
            resp[key].append(existing_map)

    return resp


def add_map(vios_w, scsi_mapping):
    """Will add the mapping to the VIOS wrapper, if not already included.

    This method has the logic in place to detect if the storage from the
    mapping is already part of a SCSI mapping.  If so, it will not re-add
    the mapping to the VIOS wrapper.

    The new mapping is added to the wrapper, but it is up to the invoker to
    call the update method on the wrapper.

    :param vios_w: The Virtual I/O Server wrapping to add the mapping to.
    :param scsi_mapping: The scsi mapping to include in the VIOS.
    :return: The scsi_mapping that was added.  None if the mapping was already
             on the vios_w.
    """
    # Check to see if the mapping is already in the system.
    lpar_uuid = util.get_req_path_uuid(scsi_mapping.client_lpar_href,
                                       preserve_case=True)
    existing_mappings = find_maps(vios_w.scsi_mappings,
                                  client_lpar_id=lpar_uuid,
                                  stg_elem=scsi_mapping.backing_storage)
    if len(existing_mappings) > 0:
        return None
    vios_w.scsi_mappings.append(scsi_mapping)
    return scsi_mapping


def remove_maps(vwrap, client_lpar_id, match_func=None, include_orphans=True):
    """Remove one or more SCSI mappings from a VIOS wrapper.

    The changes are not flushed back to the REST server.

    :param vwrap: VIOS EntryWrapper representing the Virtual I/O Server whose
                  SCSI mappings are to be updated.
    :param client_lpar_id: The integer short ID or string UUID of the client VM
    :param match_func: (Optional) Matching function suitable for passing to
                       find_maps.  See that method's match_func parameter.
                       Defaults to None (match only on client_lpar_id).
    :param include_orphans: (Optional) An "orphan" contains a server adapter
                            but no client adapter.  If this parameter is True,
                            mappings with no client adapter will be considered
                            for removal. If False, mappings with no client
                            adapter will be left alone, regardless of any other
                            criteria.  Default: True (remove orphans).
    :return: The list of removed mappings.
    """
    resp_list = []
    for matching_map in find_maps(
            vwrap.scsi_mappings, client_lpar_id=client_lpar_id,
            match_func=match_func, include_orphans=include_orphans):
        vwrap.scsi_mappings.remove(matching_map)
        resp_list.append(matching_map)
    return resp_list


def detach_storage(vwrap, client_lpar_id, match_func=None):
    """Detach the storage from all matching SCSI mappings.

    We do this by removing the Storage and TargetDevice child elements.  This
    method only updates the vwrap.  It does not POST back to the REST server.
    It does not lock.

    :param vwrap: VIOS EntryWrapper representing the Virtual I/O Server whose
                  SCSI mappings are to be updated.
    :param client_lpar_id: The integer short ID or string UUID of the client VM
    :param match_func: (Optional) Matching function suitable for passing to
                       find_maps.  See that method's match_func parameter.
                       Defaults to None (match only on client_lpar_id).
    :return: The list of SCSI mappings which were modified, in their original
             (storage-attached) form.
    """
    # Rather than modifying the matching mappings themselves, we remove them
    # and recreate them without storage.
    resp_list = []
    for match in find_maps(
            vwrap.scsi_mappings, client_lpar_id=client_lpar_id,
            match_func=match_func, include_orphans=True):
        vwrap.scsi_mappings.remove(match)
        resp_list.append(match)
        vwrap.scsi_mappings.append(
            pvm_vios.VSCSIMapping.bld_from_existing(match, None))
    return resp_list


@lock.synchronized('vscsi_mapping')
@pvm_retry.retry(tries=60, argmod_func=_argmod,
                 delay_func=pvm_retry.STEPPED_RANDOM_DELAY)
def _remove_storage_elem(adapter, vios, client_lpar_id, match_func):
    """Removes the storage element from a SCSI bus and clears out bus.

    Will remove the vSCSI Mappings from the VIOS if the match_func indicates
    that the mapping is a match.  The match_func is only invoked if the
    client_lpar_id matches.

    :param adapter: The pypowervm adapter for API communication.
    :param vios: The virtual I/O server from which the mapping should be
                 removed.  This may be the VIOS's UUID string OR an existing
                 VIOS EntryWrapper.  If the latter, it must have been retrieved
                 using the VIO_SMAP extended attribute group.
    :param client_lpar_id: The integer short ID or string UUID of the client VM
    :param match_func: Matching function suitable for passing to find_maps.
                       See that method's match_func parameter.
    :return: The VIOS wrapper representing the updated Virtual I/O Server.
             This is current with respect to etag and SCSI mappings.
    :return: The list of the storage elements that were removed from the maps.
    """
    # If the 'vios' param is a string UUID, retrieve the VIOS wrapper.
    if not isinstance(vios, pvm_vios.VIOS):
        vios = pvm_vios.VIOS.wrap(
            adapter.read(pvm_vios.VIOS.schema_type, root_id=vios,
                         xag=[c.XAG.VIO_SMAP]))

    resp_list = remove_maps(vios, client_lpar_id, match_func=match_func)

    # Update the VIOS, but only if we actually removed mappings
    if resp_list:
        vios = vios.update()

    # return the (possibly updated) VIOS and the list of removed backing
    # storage elements.
    return vios, [rmap.backing_storage for rmap in resp_list
                  if rmap.backing_storage is not None]


def gen_match_func(wcls, name_prop='name', names=None, prefixes=None,
                   udids=None):
    """Generate a matching function for find_maps' match_func param.

    :param wcls: The Wrapper class of the object being matched.
    :param name_prop: The property of the Wrapper class on which to match.
    :param names: (Optional) A list of names to match.  If names and prefixes
                  are both None or empty, all inputs of the specified wcls will
                  be matched.
    :param prefixes: (Optional) A list of prefixes that can be specified
                     to serve as identifiers for potential matches.  Ignored
                     if names is specified.  If names and prefixes are both
                     None or empty, all inputs of the specified wcls will be
                     matched.
    :param udids: (Optional) A list of UDIDs that can be specified to serve as
                  identifiers for potential matches.  Ignored if names or
                  prefixes are specified.  If all three are None or empty, all
                  inputs of the specified wcls will be matched.
    :return: A callable matching function suitable for passing to the
             match_func parameter of the find_maps method.
    """
    def match_func(existing_elem):
        if not isinstance(existing_elem, wcls):
            return False
        if names:
            return getattr(existing_elem, name_prop) in names
        if prefixes:
            for prefix in prefixes:
                if getattr(existing_elem, name_prop).startswith(prefix):
                    return True
            # prefixes specified, but none matched
            return False
        if udids:
            return existing_elem.udid in udids
        # No names, prefixes, or UDIDs specified - hit everything
        return True
    return match_func


def find_maps(mapping_list, client_lpar_id=None, match_func=None,
              stg_elem=None, include_orphans=False):
    """Filter a list of scsi mappings by LPAR ID/UUID and a matching function.

    :param mapping_list: The mappings to filter.  Iterable of VSCSIMapping.
    :param client_lpar_id: Integer short ID or string UUID of the LPAR on the
                           client side of the mapping.  Note that the UUID form
                           relies on the presence of the client_lpar_href
                           field.  Some mappings lack this field, and would
                           therefore be ignored. If client_lpar_id is not
                           passed it will return matching mappings for all
                           the lpar_ids.
    :param match_func: Callable with the following specification:

        def match_func(storage_elem)
            param storage_elem: A backing storage element wrapper (VOpt, VDisk,
                                PV, or LU) to be analyzed.  May be None (some
                                mappings have no backing storage).
            return: True if the storage_elem's mapping should be included;
                    False otherwise.

                       If neither match_func nor stg_elem is specified, the
                       default is to match everything - that is, find_maps will
                       return all mappings for the specified client_lpar_id.
                       It is illegal to specify both match_func and stg_elem.
    :param stg_elem: Match mappings associated with a specific storage element.
                     Effectively, this generates a default match_func which
                     matches on the type and name of the storage element.
                     If neither match_func nor stg_elem is specified, the
                     default is to match everything - that is, find_maps will
                     return all mappings for the specified client_lpar_id.
                     It is illegal to specify both match_func and stg_elem.
    :param include_orphans: An "orphan" contains a server adapter but no client
                            adapter.  If this parameter is True, mappings with
                            no client adapter will still be considered for
                            inclusion.  If False, mappings with no client
                            adapter will be skipped entirely, regardless of any
                            other criteria.
    :return: A list comprising the subset of the input mapping_list whose
             client LPAR IDs match client_lpar_id and whose backing storage
             elements satisfy match_func.
    :raise ValueError: If both match_func and stg_elem are specified.
    """
    if match_func and stg_elem:
        raise ValueError(_("Must not specify both match_func and stg_elem."))
    if not match_func:
        # Default no filter
        match_func = lambda x: True
    if stg_elem:
        # Match storage element on type and name
        match_func = lambda stg_el: (
            stg_el is not None and
            stg_el.schema_type == stg_elem.schema_type and
            stg_el.name == stg_elem.name)

    is_uuid = False
    client_id = None
    if client_lpar_id:
        is_uuid, client_id = uuid.id_or_uuid(client_lpar_id)
    matching_maps = []
    for existing_scsi_map in mapping_list:
        # No client, continue on unless including orphans.
        if not include_orphans and existing_scsi_map.client_adapter is None:
            continue

        # If to a different VM, continue on.
        href = existing_scsi_map.client_lpar_href
        if is_uuid and (not href or client_id != util.get_req_path_uuid(
                href, preserve_case=True)):
            continue
        elif (client_lpar_id and not is_uuid and
              # Use the server adapter in case this is an orphan.
              existing_scsi_map.server_adapter.lpar_id != client_id):
            continue

        if match_func(existing_scsi_map.backing_storage):
            # Found a match!
            matching_maps.append(existing_scsi_map)
    return matching_maps


def index_mappings(maps):
    """Create an index dict of SCSI mappings to facilitate reverse lookups.

    :param maps: Iterable of VSCSIMapping to index.
    :return: A dict of the form:
        { 'by-lpar-id': { str(lpar_id): [VSCSIMapping, ...], ... },
          'by-lpar-uuid': { lpar_uuid: [VSCSIMapping, ...], ... },
          'by-storage-udid': { storage_udid: [VSCSIMapping, ...], ... }
        }
        ...where:
        - lpar_id is the short integer ID (not UUID) of the LPAR, stringified.
        - lpar_uuid is the UUID of the LPAR.
        - storage_udid is the Unique Device Identifier (UDID) of the backing
            Storage element associated with the mapping.

        While the outermost dict is guaranteed to have all keys, the inner
        dicts may be empty.  However, if an inner dict has a member, its list
        of mappings is guaranteed to be nonempty.
    """
    ret = {'by-lpar-id': {}, 'by-lpar-uuid': {}, 'by-storage-udid': {}}

    def add(key, ident, smap):
        """Add a mapping to an index.

        :param key: The top-level key name ('by-lpar-uuid', etc.)
        :param ident: The lower-level key name (e.g. the lpar_uuid)
        :param smap: The mapping to add to the index.
        """
        ident = str(ident)
        if not ident:
            return
        if ident not in ret[key]:
            ret[key][ident] = []
        ret[key][ident].append(smap)

    for smap in maps:
        clhref = smap.client_lpar_href
        if clhref:
            add('by-lpar-uuid',
                util.get_req_path_uuid(clhref, preserve_case=True), smap)

        clid = None
        # Mapping may not have a client adapter, but will always have a server
        # adapter - so get the LPAR ID from the server adapter.
        if smap.server_adapter:
            clid = smap.server_adapter.lpar_id
        add('by-lpar-id', clid, smap)

        stg = smap.backing_storage
        if stg:
            add('by-storage-udid', stg.udid, smap)

    return ret


def remove_vopt_mapping(adapter, vios, client_lpar_id, media_name=None,
                        udid=None):
    """Will remove the mapping for VOpt media.

    This method will remove the mapping between the virtual optical media
    and the client partition.  It does not delete the virtual optical media.
    Will leave other elements on the vSCSI bus intact.

    :param adapter: The pypowervm adapter for API communication.
    :param vios: The virtual I/O server from which the mapping should be
                 removed.  This may be the VIOS's UUID string OR an existing
                 VIOS EntryWrapper.  If the latter, it must have been retrieved
                 using the VIO_SMAP extended attribute group.
    :param client_lpar_id: The integer short ID or string UUID of the client VM
    :param media_name: (Optional) The name of the virtual optical media to
                       remove from the SCSI bus.  If both media_name and udid
                       are None, will remove all virtual optical media mappings
                       associated with the specified client_lpar_id
    :param udid: (Optional) The UDID of the virtual optical media to remove
                 from the SCSI bus.  Ignored if media_name is specified.  If
                 both media_name and udid are None, will remove all virtual
                 optical media mappings associated with the client_lpar_id.
    :return: The VIOS wrapper representing the updated Virtual I/O Server.
             This is current with respect to etag and SCSI mappings.
    :return: A list of the backing VOpt media that was removed.
    """
    names = [media_name] if media_name else None
    udids = [udid] if udid else None
    return _remove_storage_elem(
        adapter, vios, client_lpar_id, gen_match_func(
            pvm_stor.VOptMedia, name_prop='media_name', names=names,
            udids=udids))


def remove_vdisk_mapping(adapter, vios, client_lpar_id, disk_names=None,
                         disk_prefixes=None, udids=None):
    """Will remove the mapping for VDisk media.

    This method will remove the mapping between the virtual disk and the
    client partition.  It does not delete the virtual disk.  Will leave other
    elements on the vSCSI bus intact.

    :param adapter: The pypowervm adapter for API communication.
    :param vios: The virtual I/O server from which the mapping should be
                 removed.  This may be the VIOS's UUID string OR an existing
                 VIOS EntryWrapper.  If the latter, it must have been retrieved
                 using the VIO_SMAP extended attribute group.
    :param client_lpar_id: The integer short ID or string UUID of the client VM
    :param disk_names: (Optional) A list of names of the virtual disk to remove
                       from the SCSI bus.  If disk_names, disk_prefixes, and
                       udids are all None/empty, will remove all virtual disk
                       mappings associated with the specified client_lpar_id.
    :param disk_prefixes: (Optional) A list of prefixes that can be specified
                          to serve as identifiers for potential disks.  Ignored
                          if disk_names is specified.  If disk_names,
                          disk_prefixes, and udids are all None/empty, will
                          remove all virtual disk mappings associated with the
                          specified client_lpar_id.
    :param udids: (Optional) A list of UDIDs of the virtual disks to remove
                  from the SCSI bus.  Ignored if disk_names or disk_prefixes
                  are specified.  If all three are None/empty, will remove all
                  virtual disk mappings associated with the specified
                  client_lpar_id.
    :return: The VIOS wrapper representing the updated Virtual I/O Server.
             This is current with respect to etag and SCSI mappings.
    :return: A list of the backing VDisk objects that were removed.
    """

    return _remove_storage_elem(
        adapter, vios, client_lpar_id, gen_match_func(
            pvm_stor.VDisk, names=disk_names, prefixes=disk_prefixes,
            udids=udids))


def remove_lu_mapping(adapter, vios, client_lpar_id, disk_names=None,
                      disk_prefixes=None, udids=None):
    """Remove mappings for one or more SSP LUs associated with an LPAR.

    This method will remove the mapping between the Logical Unit and the
    client partition.  It does not delete the LU.  Will leave other elements on
    the vSCSI bus intact.

    :param adapter: The pypowervm adapter for API communication.
    :param vios: The virtual I/O server from which the mapping should be
                 removed.  This may be the VIOS's UUID string OR an existing
                 VIOS EntryWrapper.  If the latter, it must have been retrieved
                 using the VIO_SMAP extended attribute group.
    :param client_lpar_id: The integer short ID or string UUID of the client VM
    :param disk_names: (Optional) A list of names of the LUs to remove from
                       the SCSI bus.  If disk_names, disk_prefixes, and
                       udids are all None/empty, will remove all logical unit
                       mappings associated with the specified client_lpar_id.
    :param disk_prefixes: (Optional) A list of prefixes that can be specified
                          to serve as identifiers for potential disks.  Ignored
                          if disk_names is specified.  If disk_names,
                          disk_prefixes, and udids are all None/empty, will
                          remove all logical unit mappings associated with the
                          specified client_lpar_id.
    :param udids: (Optional) A list of UDIDs of the logical units to remove
                  from the SCSI bus.  Ignored if disk_names or disk_prefixes
                  are specified.  If all three are None/empty, will remove all
                  logical unit mappings associated with the specified
                  client_lpar_id.
    :return: The VIOS wrapper representing the updated Virtual I/O Server.
             This is current with respect to etag and SCSI mappings.
    :return: A list of LU EntryWrappers representing the mappings that were
             removed.
    """

    return _remove_storage_elem(
        adapter, vios, client_lpar_id, gen_match_func(
            pvm_stor.LU, names=disk_names, prefixes=disk_prefixes,
            udids=udids))


def remove_pv_mapping(adapter, vios, client_lpar_id, backing_dev, udid=None):
    """Will remove the PV mapping.

    This method will remove the pv mapping. It does not delete the device.
    Will leave other elements on the vSCSI bus intact.

    :param adapter: The pypowervm adapter for API communication.
    :param vios: The virtual I/O server from which the mapping should be
                 removed.  This may be the VIOS's UUID string OR an existing
                 VIOS EntryWrapper.  If the latter, it must have been retrieved
                 using the VIO_SMAP extended attribute group.
    :param client_lpar_id: The integer short ID or string UUID of the client VM
    :param backing_dev: The physical volume name to be removed.  If both
                        backing_dev and udid are None, will remove all physical
                        volume mappings associated with the specfied
                        client_lpar_id.
    :param udid: (Optional) UDID of the physical volume to remove from the SCSI
                  bus.  Ignored if backing_dev is not None.  If backing_dev and
                  udid are both None, will remove all physical volume mappings
                  associated with the specified client_lpar_id.
    :return: The VIOS wrapper representing the updated Virtual I/O Server.
             This is current with respect to etag and SCSI mappings.
    :return: A list of the backing physical device objects that were removed.
    """
    names = [backing_dev] if backing_dev else None
    udids = [udid] if udid else None
    return _remove_storage_elem(
        adapter, vios, client_lpar_id, gen_match_func(
            pvm_stor.PV, names=names, udids=udids))



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the raw LPAR long term metrics."""

import json
from pypowervm.tests.test_utils import pvmhttp
from pypowervm.wrappers.pcm import lpar as pcm_lpar
import testtools


LPAR_DATA = 'lpar_pcm_data.txt'


class TestLparLTM(testtools.TestCase):

    def setUp(self):
        super(TestLparLTM, self).setUp()
        self.raw_json = pvmhttp.PVMFile(LPAR_DATA).body

    def test_parse(self):
        info = pcm_lpar.LparInfo(self.raw_json)
        self.assertIsNotNone(info)
        # Validate the Lpar metrics.
        # There are metrics for four Lpars.
        self.assertEqual(6, len(info.lpars_util))
        # Get the first Lpar and assert its metrics
        lpar = info.lpars_util[0]
        self.assertEqual("Ubuntu1410", lpar.name)
        self.assertIsNotNone(lpar.memory)
        self.assertEqual(80, lpar.memory.pct_real_mem_avbl)
        self.assertEqual(1024, lpar.memory.total_pg_count)
        self.assertEqual(512, lpar.memory.free_pg_count)
        self.assertEqual(64, lpar.memory.active_pg_count)
        self.assertEqual(1048576, lpar.memory.real_mem_size_bytes)
        self.assertEqual(61, lpar.memory.pct_real_mem_free)
        self.assertEqual(25, lpar.memory.vm_pg_out_rate)
        # Get 3rd(random) VM and assert its metrics
        lpar = info.lpars_util[2]
        self.assertEqual("test_vm3", lpar.name)
        self.assertIsNotNone(lpar.memory)
        self.assertEqual(82, lpar.memory.pct_real_mem_avbl)
        self.assertEqual(4096, lpar.memory.total_pg_count)
        self.assertEqual(2048, lpar.memory.free_pg_count)
        self.assertEqual(256, lpar.memory.active_pg_count)
        self.assertEqual(1048576, lpar.memory.real_mem_size_bytes)
        self.assertEqual(60, lpar.memory.pct_real_mem_free)
        self.assertEqual(0, lpar.memory.vm_pg_out_rate)
        # Assert that we have entries in JSON for VMs which were in error
        metric_json = json.loads(self.raw_json)
        self.assertEqual("3B0237F9-26F1-41C7-BE57-A08C9452AD9D",
                         metric_json['lparUtil'][4]['name'])
        self.assertEqual("vm_inactive_rmc",
                         metric_json['lparUtil'][5]['name'])
        # Assert that powered off VM has 100 percent free memory.
        lpar = info.lpars_util[4]
        self.assertEqual("3B0237F9-26F1-41C7-BE57-A08C9452AD9D", lpar.name)
        self.assertIsNotNone(lpar.memory)
        self.assertIsNone(lpar.memory.pct_real_mem_avbl)
        self.assertIsNone(lpar.memory.total_pg_count)
        self.assertIsNone(lpar.memory.free_pg_count)
        self.assertIsNone(lpar.memory.active_pg_count)
        self.assertIsNone(lpar.memory.real_mem_size_bytes)
        self.assertEqual(100, lpar.memory.pct_real_mem_free)
        # Assert that LPAR with inactive RMC has no free memory.
        lpar = info.lpars_util[5]
        self.assertEqual("vm_inactive_rmc", lpar.name)
        self.assertIsNotNone(lpar.memory)
        self.assertIsNone(lpar.memory.pct_real_mem_avbl)
        self.assertIsNone(lpar.memory.total_pg_count)
        self.assertIsNone(lpar.memory.free_pg_count)
        self.assertIsNone(lpar.memory.active_pg_count)
        self.assertIsNone(lpar.memory.real_mem_size_bytes)
        self.assertEqual(0, lpar.memory.pct_real_mem_free)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Construction and basic validation of an LPAR or VIOS EntryWrapper."""

import abc
import six

from oslo_log import log as logging

from pypowervm import i18n
from pypowervm.wrappers import base_partition as bp
from pypowervm.wrappers import logical_partition as lpar
from pypowervm.wrappers import virtual_io_server as vios

# Dict keys used for input to the builder
NAME = 'name'
ENV = 'env'
UUID = 'uuid'
ID = 'id'

MEM = 'memory'
MAX_MEM = 'max_mem'
MIN_MEM = 'min_mem'
AME_FACTOR = 'ame_factor'

DED_PROCS = 'dedicated_proc'
VCPU = 'vcpu'
MAX_VCPU = 'max_vcpu'
MIN_VCPU = 'min_vcpu'
DED_PROC_KEYS = ()

PROC_UNITS = 'proc_units'
MAX_PROC_U = 'max_proc_units'
MIN_PROC_U = 'min_proc_units'
PROC_UNITS_KEYS = (PROC_UNITS, MAX_PROC_U, MIN_PROC_U)

SHARING_MODE = 'sharing_mode'
UNCAPPED_WEIGHT = 'uncapped_weight'
SPP = 'proc_pool'
MAX_IO_SLOTS = 'max_io_slots'
AVAIL_PRIORITY = 'avail_priority'
SRR_CAPABLE = 'srr_capability'
PROC_COMPAT = 'processor_compatibility'

# IBMi specific keys
ALT_LOAD_SRC = 'alt_load_src'
CONSOLE = 'console'
LOAD_SRC = 'load_src'
RESTRICTED_IO = 'restricted_io'

# The minimum attributes that must be supplied to create an LPAR
MINIMUM_ATTRS = (NAME, MEM, VCPU)
# Keys that indicate that shared processors are being configured
SHARED_PROC_KEYS = (PROC_UNITS_KEYS + (UNCAPPED_WEIGHT,))

MEM_LOW_BOUND = 128
VCPU_LOW_BOUND = 1
PROC_UNITS_LOW_BOUND = 0.05
MAX_LPAR_NAME_LEN = 31

# Defaults
DEF_PROC_UNIT_FACT = 0.5
DEF_MAX_SLOT = 64
DEF_UNCAPPED_WT = 64
DEF_SPP = 0
DEF_AVAIL_PRI = 127
DEF_SRR = 'false'

LOG = logging.getLogger(__name__)

# TODO(IBM) translation
_LE = i18n._


class LPARBuilderException(Exception):
    """Exceptions thrown from the lpar builder."""
    pass


@six.add_metaclass(abc.ABCMeta)
class Standardize(object):
    """Interface class to standardize the LPAR definition

    A standardizer is responsible for validating the LPAR attributes
    that are presented and augmenting those which are required to create
    the LPAR.
    """
    def __init__(self):
        self.attr = None

    def set_attr(self, attr):
        """Set the attributes to be validated and standardized.

        :param attr: dict of lpar attributes provided by the user
        """
        self.attr = attr

    def general(self):
        """Validates and standardizes the general LPAR attributes.

        :returns: dict of attributes.
            Expected: NAME, ENV, MAX_IO_SLOTS, AVAIL_PRIORITY,
                      PROC_COMPAT
            Optional: SRR_CAPABLE, UUID, ID
                      IBMi value: CONSOLE, LOAD_SRC, ALT_LOAD_SRC,
                                  RESTRICTED_IO
        """
        pass

    def memory(self):
        """Validates and standardizes the memory LPAR attributes.

        :returns: dict of attributes.
            Expected: MEM, MIN_MEM, MAX_MEM
        """
        pass

    def ded_proc(self):
        """Validates and standardizes the dedicated processor LPAR attributes.

        :returns: dict of attributes.
            Expected: VCPU, MIN_VCPU, MAX_VCPU, SHARING_MODE
        """
        pass

    def shr_proc(self):
        """Validates and standardizes the shared processor LPAR attributes.

        :returns: dict of attributes.
            Expected: VCPU, MIN_VCPU, MAX_VCPU, PROC_UNITS, MAX_PROC_U,
                      MIN_PROC_U, SHARING_MODE,
                      UNCAPPED_WEIGHT(if UNCAPPED)
        """
        pass


class DefaultStandardize(Standardize):
    """Default standardizer.

    This class implements the Standardizer interface.  It takes a
    simple approach for augmenting missing LPAR settings.  It does
    reasonable validation of the LPAR attributes.

    It first validates the user input as-is, then fills in any missing
    attributes that are required and supported by the host.  Finally,
    it validates what it's sending back to the caller.  If any validation
    rules are missed, the PowerVM management interface will catch them
    and surface an error at that time.

    """
    def __init__(self, mngd_sys,
                 proc_units_factor=DEF_PROC_UNIT_FACT, max_slots=DEF_MAX_SLOT,
                 uncapped_weight=DEF_UNCAPPED_WT, spp=DEF_SPP,
                 avail_priority=DEF_AVAIL_PRI, srr=DEF_SRR,
                 proc_compat=bp.LPARCompat.DEFAULT):
        """Initialize the standardizer

        :param mngd_sys: managed_system wrapper of the host to deploy to.
            This is used to validate the fields and standardize against the
            host.
        :param proc_units_factor: amount of proc units to assign to each vcpu
            if proc units are not specified
        :param max_slots: number of max io slots to assign, if not specified
        :param uncapped_weight: the uncapped weight to use if the processors
            are shared and a weight is not specified
        :param spp: shared processor pool to assign if the processors are
            shared and the pool is not specified
        :param avail_priority: availability priority of the LPAR
        :param srr: simplified remote restart capable
        :param proc_compat: processor compatibility mode value
        """

        super(DefaultStandardize, self).__init__()
        self.mngd_sys = mngd_sys
        self.proc_units_factor = proc_units_factor
        if proc_units_factor > 1 or proc_units_factor < 0.05:
            msg = _LE('Processor units factor must be between 0.05 and 1.0. '
                      'Value: %s') % proc_units_factor
            raise LPARBuilderException(msg)
        self.max_slots = max_slots
        self.uncapped_weight = uncapped_weight
        self.spp = spp
        self.avail_priority = avail_priority
        self.srr = srr
        self.proc_compat = proc_compat

    def _set_prop(self, attr, prop, base_prop, convert_func=str):
        """Copies a property if present or copies the base property."""
        attr[prop] = convert_func(self.attr.get(prop, self.attr[base_prop]))

    def _set_val(self, attr, prop, value=None, convert_func=str):
        """Copies a property if present or uses the supplied value."""
        val = self.attr.get(prop, value)
        if val is not None:
            attr[prop] = convert_func(val)

    def _validate_general(self, attrs=None, partial=False):
        if attrs is None:
            attrs = self.attr
        name_len = len(attrs[NAME])
        if name_len < 1 or name_len > MAX_LPAR_NAME_LEN:

            msg = _LE("Logical partition name has invalid length."
                      " Name: %s") % attrs[NAME]
            raise LPARBuilderException(msg)
        LPARType(attrs.get(ENV), allow_none=partial).validate()
        IOSlots(attrs.get(MAX_IO_SLOTS), allow_none=partial).validate()
        AvailPriority(attrs.get(AVAIL_PRIORITY), allow_none=partial).validate()
        IDBoundField(attrs.get(ID), allow_none=True).validate()
        # SRR is always optional since the host may not be capable of it.
        SimplifiedRemoteRestart(attrs.get(SRR_CAPABLE),
                                allow_none=True).validate()
        ProcCompatMode(attrs.get(PROC_COMPAT),
                       host_modes=self.mngd_sys.proc_compat_modes,
                       allow_none=partial).validate()

        # Validate fields specific to IBMi
        if attrs.get(ENV, '') == bp.LPARType.OS400:
            RestrictedIO(attrs.get(RESTRICTED_IO), allow_none=True).validate()

    def _validate_memory(self, attrs=None, partial=False):
        if attrs is None:
            attrs = self.attr
        host_ame_cap = self.mngd_sys.get_capabilities()[
            'active_memory_expansion_capable']
        mem = Memory(attrs.get(MIN_MEM), attrs.get(MEM), attrs.get(MAX_MEM),
                     attrs.get(AME_FACTOR), host_ame_cap,
                     self.mngd_sys.memory_region_size, allow_none=partial)
        mem.validate()

    def _validate_shared_proc(self, attrs=None, partial=False):
        if attrs is None:
            attrs = self.attr

        # Validate the vcpu first
        VCpu(attrs.get(MIN_VCPU), attrs.get(VCPU), attrs.get(MAX_VCPU),
             partial).validate()
        # Validate the proc units
        ProcUnits(attrs.get(MIN_PROC_U), attrs.get(PROC_UNITS),
                  attrs.get(MAX_PROC_U), allow_none=partial).validate()
        # TODO(IBM): Validate any shared CPU associated parameters

    def _validate_lpar_ded_cpu(self, attrs=None, partial=False):
        if attrs is None:
            attrs = self.attr

        VCpu(attrs.get(MIN_VCPU), attrs.get(VCPU), attrs.get(MAX_VCPU),
             allow_none=partial).validate()

        # If specified, ensure the dedicated procs value is valid
        DedicatedProc(attrs.get(DED_PROCS), allow_none=True).validate()
        DedProcShareMode(attrs.get(SHARING_MODE), allow_none=True).validate()

    def general(self):
        # Validate the settings sent in
        self._validate_general(partial=True)

        bld_attr = {NAME: self.attr[NAME]}
        self._set_val(bld_attr, ID, convert_func=int)
        self._set_val(bld_attr, UUID)
        self._set_val(bld_attr, ENV, bp.LPARType.AIXLINUX,
                      convert_func=LPARType.convert_value)
        self._set_val(bld_attr, MAX_IO_SLOTS, self.max_slots)
        self._set_val(bld_attr, AVAIL_PRIORITY, self.avail_priority)
        # See if the host is capable of SRR before setting it.
        host_cap = self.mngd_sys.get_capabilities()
        if host_cap['simplified_remote_restart_capable']:
            self._set_val(bld_attr, SRR_CAPABLE, self.srr,
                          convert_func=SimplifiedRemoteRestart.convert_value)
        self._set_val(bld_attr, PROC_COMPAT, bp.LPARCompat.DEFAULT,
                      convert_func=ProcCompatMode.convert_value)

        # Build IBMi attributes
        if bld_attr[ENV] == bp.LPARType.OS400:
            self._set_val(bld_attr, CONSOLE, value='HMC')
            self._set_val(bld_attr, LOAD_SRC, value='0')
            self._set_val(bld_attr, ALT_LOAD_SRC, value='NONE')
            if host_cap['ibmi_restrictedio_capable']:
                self._set_val(bld_attr, RESTRICTED_IO, value=True,
                              convert_func=RestrictedIO.convert_value)

        # Validate the attributes
        self._validate_general(attrs=bld_attr)
        return bld_attr

    def memory(self):
        # Validate the partial settings
        self._validate_memory(partial=True)

        bld_attr = {MEM: self.attr[MEM]}
        self._set_prop(bld_attr, MAX_MEM, MEM)
        self._set_prop(bld_attr, MIN_MEM, MEM)

        # Validate the full memory settings
        self._validate_memory(attrs=bld_attr)
        return bld_attr

    def shr_proc(self):
        def _compare(prop, value, compare_func, typ):
            v1 = self.attr.get(prop)
            # Ensure the property is specified
            if v1 is None:
                return value

            # Compare
            return compare_func(typ(v1), value)

        # Validate the partial settings
        self._validate_shared_proc(partial=True)

        bld_attr = {VCPU: self.attr[VCPU]}
        self._set_prop(bld_attr, MAX_VCPU, VCPU)
        self._set_prop(bld_attr, MIN_VCPU, VCPU)

        # See if we need to calculate a default proc_units value and min/max
        # Before setting the proc units ensure it's between min/max
        spec_proc_units = self.attr.get(PROC_UNITS)
        if spec_proc_units is None:
            proc_units = int(bld_attr[VCPU]) * self.proc_units_factor

            # Ensure it's at least as large as a specified min value
            proc_units = _compare(MIN_PROC_U, proc_units, max, float)

            # Ensure it's smaller than a specified max value
            proc_units = _compare(MAX_PROC_U, proc_units, min, float)
        else:
            proc_units = float(spec_proc_units)

        self._set_val(bld_attr, PROC_UNITS, proc_units)
        self._set_val(bld_attr, MIN_PROC_U, proc_units)
        self._set_val(bld_attr, MAX_PROC_U, proc_units)
        self._set_val(bld_attr, SHARING_MODE, bp.SharingMode.UNCAPPED)

        # If uncapped sharing mode then set the weight
        if bld_attr.get(SHARING_MODE) == bp.SharingMode.UNCAPPED:
            self._set_val(bld_attr, UNCAPPED_WEIGHT, self.uncapped_weight)
        self._set_val(bld_attr, SPP, self.spp)

        # Validate all the values
        self._validate_shared_proc(attrs=bld_attr)
        return bld_attr

    def ded_proc(self):
        self._validate_lpar_ded_cpu(partial=True)
        # Set the proc based on vcpu field
        bld_attr = {VCPU: self.attr[VCPU]}
        self._set_prop(bld_attr, MAX_VCPU, VCPU)
        self._set_prop(bld_attr, MIN_VCPU, VCPU)
        self._set_val(bld_attr, SHARING_MODE,
                      bp.DedicatedSharingMode.SHARE_IDLE_PROCS,
                      convert_func=DedProcShareMode.convert_value)
        self._validate_lpar_ded_cpu(attrs=bld_attr)
        return bld_attr


@six.add_metaclass(abc.ABCMeta)
class Field(object):
    """Represents a field to validate."""
    _type = str

    def __init__(self, value, name=None, allow_none=True):
        self.name = name if name is not None else self.__class__._name
        self.value = value
        self.typed_value = None
        self.allow_none = allow_none

    @classmethod
    def convert_value(cls, value):
        """Static converter for the Field type."""
        return cls._type(value)

    def _type_error(self, value, exc=TypeError):
        values = dict(field=self.name, value=value)
        msg = _LE("Field '%(field)s' has invalid value: '%(value)s'") % values
        LOG.error(msg)
        raise exc(msg)

    def _convert_value(self, value):
        """Does the actual conversion of the value and returns it."""
        try:
            return self.convert_value(value)
        except (TypeError, ValueError) as e:
            self._type_error(value, exc=e.__class__)

    def _convert(self):
        """Converts the value and saves it away for future use."""
        self.typed_value = self._convert_value(self.value)

    def validate(self):
        # Check if the value is none and we allow that
        if self.value is None:
            if not self.allow_none:
                self._type_error(None)
        else:
            # The base value is not none, so see if we should convert it
            if self.typed_value is None:
                self._convert()


@six.add_metaclass(abc.ABCMeta)
class BoolField(Field):
    """Facilitates validating boolean fields."""
    _type = bool

    @classmethod
    def convert_value(cls, value):
        # Special case string values, so random strings don't map to True
        if isinstance(value, six.string_types):
            if value.lower() in ['true', 'yes']:
                return True
            if value.lower() in ['false', 'no']:
                return False
        elif isinstance(value, bool):
                return value
        raise ValueError('Could not convert %s.' % value)


@six.add_metaclass(abc.ABCMeta)
class ChoiceField(Field):
    _choices = None

    @classmethod
    def convert_value(cls, value):
        return cls._validate_choices(value, cls._choices)

    @classmethod
    def _validate_choices(cls, value, choices):
        if value is None:
            raise ValueError(_LE('None value is not valid.'))
        for choice in choices:
            if value.lower() == choice.lower():
                return choice
        # If we didn't find it, that's a problem...
        values = dict(value=value, field=cls._name,
                      choices=choices)
        msg = _LE("Value '%(value)s' is not valid "
                  "for field '%(field)s' with acceptable "
                  "choices: %(choices)s") % values
        raise ValueError(msg)

    def validate(self):
        if self.value is None and self.allow_none:
            return
        super(ChoiceField, self).validate()
        self._validate_choices(self.value, self._choices)


@six.add_metaclass(abc.ABCMeta)
class BoundField(Field):
    _min_bound = None
    _max_bound = None

    def validate(self):
        super(BoundField, self).validate()
        # If value was not converted to the type, then don't validate bounds
        if self.typed_value is None:
            return
        if (self._min_bound is not None and
                self.typed_value < self._convert_value(self._min_bound)):
            values = dict(field=self.name, value=self.typed_value,
                          minimum=self._min_bound)
            msg = _LE("Field '%(field)s' has a value below the minimum. "
                      "Value: %(value)s Minimum: %(minimum)s") % values
            LOG.error(msg)
            raise ValueError(msg)

        if (self._max_bound is not None and
                self.typed_value > self._convert_value(self._max_bound)):
            values = dict(field=self.name, value=self.typed_value,
                          maximum=self._max_bound)
            msg = _LE("Field '%(field)s' has a value above the maximum. "
                      "Value: %(value)s Maximum: %(maximum)s") % values

            LOG.error(msg)
            raise ValueError(msg)


@six.add_metaclass(abc.ABCMeta)
class IntBoundField(BoundField):
    _type = int


@six.add_metaclass(abc.ABCMeta)
class FloatBoundField(BoundField):
    _type = float


@six.add_metaclass(abc.ABCMeta)
class MinDesiredMaxField(object):

    def __init__(self, field_type, min_name, des_name, max_name,
                 min_value, desired_value, max_value,
                 min_min=None, max_max=None, name=None,
                 allow_none=True):
        self.name = name if name is not None else self.__class__._name

        self.min_field = field_type(
            min_value, name=min_name, allow_none=allow_none)
        self.min_field._max_bound = desired_value
        self.min_field._min_bound = min_min

        self.des_field = field_type(
            desired_value, name=des_name, allow_none=allow_none)
        self.des_field._min_bound = min_value
        self.des_field._max_bound = max_value

        self.max_field = field_type(
            max_value, name=max_name, allow_none=allow_none)
        self.max_field._min_bound = desired_value
        self.max_field._max_bound = max_max

    def validate(self):
        # Do specific validations before the general ones
        for fld in [self.min_field, self.des_field, self.max_field]:
            if fld.value is not None or not fld.allow_none:
                fld._convert()

        # Ensure the desired value is between the min and max
        if (self.des_field.typed_value and self.max_field.typed_value and
                self.des_field.typed_value > self.max_field.typed_value):
            values = dict(desired_field=self.des_field.name,
                          max_field=self.max_field.name,
                          desired=self.des_field.typed_value,
                          maximum=self.max_field.typed_value)
            msg = _LE("The '%(desired_field)s' has a value above the "
                      "'%(max_field)s' value. "
                      "Desired: %(desired)s Maximum: %(maximum)s") % values

            LOG.error(msg)
            raise ValueError(msg)
        # Now the minimum
        if (self.des_field.typed_value and self.min_field.typed_value and
                self.des_field.typed_value < self.min_field.typed_value):
            values = dict(desired_field=self.des_field.name,
                          min_field=self.min_field.name,
                          desired=self.des_field.typed_value,
                          minimum=self.min_field.typed_value)
            msg = _LE("The '%(desired_field)s' has a value below the "
                      "'%(min_field)s' value. "
                      "Desired: %(desired)s Minimum: %(minimum)s") % values

            LOG.error(msg)
            raise ValueError(msg)

        # Now the fields individually
        self.min_field.validate()
        self.des_field.validate()
        self.max_field.validate()


class Memory(MinDesiredMaxField):
    _name = 'Memory'

    def __init__(self, min_value, desired_value, max_value,
                 ame_ef, host_ame_cap, lmb_size, allow_none=True):
        super(Memory, self).__init__(
            IntBoundField, 'Minimum Memory', 'Desired Memory',
            'Maximum Memory', min_value, desired_value, max_value,
            allow_none=allow_none)
        self.lmb_size = lmb_size
        # Set the lowest memory we'll honor
        self.min_field._min_bound = MEM_LOW_BOUND
        # Don't allow the desired memory to not be specified.
        self.des_field.allow_none = False
        self.ame_ef = ame_ef
        self.host_ame_cap = host_ame_cap

    def validate(self):
        super(Memory, self).validate()
        self._validate_lmb_size()
        self._validate_ame()

    def _validate_lmb_size(self):
        # Validate against the LMB size
        if self.lmb_size is not None:
            # For each value, make sure it's a multiple
            for x in [self.min_field.typed_value,
                      self.des_field.typed_value,
                      self.max_field.typed_value]:
                if x is not None and (x % self.lmb_size) != 0:
                    values = dict(lmb_size=self.lmb_size, value=x)
                    msg = _LE("Memory value is not a multiple of the "
                              "logical memory block size (%(lmb_size)s) of "
                              " the host.  Value: %(value)s") % values
                    raise ValueError(msg)

    def _validate_ame(self):
        # Validate the expansion factor value
        if self.ame_ef is not None:
            exp_fact_float = round(float(self.ame_ef), 2)
            values = dict(value=self.ame_ef)
            if not self.host_ame_cap and exp_fact_float != 0:
                msg = _LE("The managed system does not support active memory "
                          "expansion. The expansion factor value '%(value)s' "
                          "is not valid.") % values
                raise ValueError(msg)
            if (exp_fact_float != 0 and exp_fact_float < 1 or
                    exp_fact_float > 10):
                msg = _LE("Active memory expansion value must be greater "
                          "than or equal to 1.0 and less than or equal to "
                          "10.0. A value of 0 is also valid and indicates "
                          "that AME is off. '%(value)s' is not "
                          "valid.") % values
                raise ValueError(msg)


class VCpu(MinDesiredMaxField):
    _name = 'VCPU'

    def __init__(self, min_value, desired_value, max_value, allow_none=True):
        super(VCpu, self).__init__(
            IntBoundField, 'Minimum VCPU', 'Desired VCPU',
            'Maximum VCPU', min_value, desired_value, max_value,
            allow_none=allow_none)
        # Set the lowest VCPU we'll honor
        self.min_field._min_bound = VCPU_LOW_BOUND


class ProcUnits(MinDesiredMaxField):
    _name = 'ProcUnits'

    def __init__(self, min_value, desired_value, max_value, allow_none=True):
        super(ProcUnits, self).__init__(
            FloatBoundField, 'Minimum Proc Units',
            'Desired Proc Units', 'Maximum Proc Units', min_value,
            desired_value, max_value,
            allow_none=allow_none)
        # Set the lowest ProcUnits we'll honor
        self.min_field._min_bound = PROC_UNITS_LOW_BOUND


class DedicatedProc(BoolField):
    _name = 'Dedicated Processors'


class LPARType(ChoiceField):
    _choices = (bp.LPARType.AIXLINUX, bp.LPARType.OS400, bp.LPARType.VIOS)
    _name = 'Logical Partition Type'

    def __init__(self, value, allow_none=False):
        super(LPARType, self).__init__(value, allow_none=allow_none)


class ProcCompatMode(ChoiceField):
    _choices = bp.LPARCompat.ALL_VALUES
    _name = 'Processor Compatability Mode'

    def __init__(self, value, host_modes=None, allow_none=True):
        super(ProcCompatMode, self).__init__(value, allow_none=allow_none)
        if host_modes:
            self._choices = host_modes


class DedProcShareMode(ChoiceField):
    _choices = bp.DedicatedSharingMode.ALL_VALUES
    _name = 'Dedicated Processor Sharing Mode'

    def __init__(self, value, allow_none=False):
        super(DedProcShareMode, self).__init__(value, allow_none=allow_none)


class IOSlots(IntBoundField):
    _min_bound = 2  # slot 0 & 1 are always in use
    _max_bound = 65534
    _name = 'I/O Slots'

    def __init__(self, value, allow_none=False):
        super(IOSlots, self).__init__(value, allow_none=allow_none)


class AvailPriority(IntBoundField):
    _min_bound = 0
    _max_bound = 255
    _name = 'Availability Priority'


class IDBoundField(IntBoundField):
    _min_bound = 1
    _name = 'ID'


class SimplifiedRemoteRestart(BoolField):
    _name = 'Simplified Remote Restart'


class RestrictedIO(BoolField):
    _name = 'Restricted IO'


class LPARBuilder(object):
    def __init__(self, adapter, attr, stdz):
        self.adapter = adapter
        self.attr = attr
        self.stdz = stdz
        for val in MINIMUM_ATTRS:
            if self.attr.get(val) is None:
                raise LPARBuilderException('Missing required attribute: %s'
                                           % val)
        stdz.set_attr(attr)

    def build_ded_proc(self):
        # Ensure no shared proc keys are present
        # TODO(IBM):

        std = self.stdz.ded_proc()
        dproc = bp.PartitionProcessorConfiguration.bld_dedicated(
            self.adapter, std[VCPU], min_proc=std[MIN_VCPU],
            max_proc=std[MAX_VCPU], sharing_mode=std[SHARING_MODE])
        return dproc

    def build_shr_proc(self):
        # Ensure no dedicated proc keys are present
        # TODO(IBM):

        std = self.stdz.shr_proc()
        # The weight may not be set if it's not uncapped
        uncapped_weight = std.get(UNCAPPED_WEIGHT)
        # Build the shared procs
        shr_proc = bp.PartitionProcessorConfiguration.bld_shared(
            self.adapter, std[PROC_UNITS], std[VCPU],
            sharing_mode=std[SHARING_MODE], uncapped_weight=uncapped_weight,
            min_proc_unit=std[MIN_PROC_U], max_proc_unit=std[MAX_PROC_U],
            min_proc=std[MIN_VCPU], max_proc=std[MAX_VCPU], proc_pool=std[SPP])
        return shr_proc

    def build_mem(self):
        std = self.stdz.memory()
        mem_wrap = bp.PartitionMemoryConfiguration.bld(
            self.adapter, std[MEM], min_mem=std[MIN_MEM], max_mem=std[MAX_MEM])
        # Determine AME enabled boolean value from expansion factor value
        if self.attr.get(AME_FACTOR) is not None:
            exp_fact_float = round(float(self.attr.get(AME_FACTOR)), 2)
            mem_wrap.exp_factor = exp_fact_float
        return mem_wrap

    def _shared_proc_keys_specified(self):
        # Check for any shared proc keys
        for key in SHARED_PROC_KEYS:
            if self.attr.get(key, None) is not None:
                return True

        # Check the sharing mode values if any
        smode = self.attr.get(SHARING_MODE, None)
        if (smode is not None and
                smode in bp.SharingMode.ALL_VALUES):
            return True

        return False

    def _dedicated_proc_keys_specified(self):
        # Check for dedicated proc keys
        # TODO(IBM):

        # Check for dedicated sharing mode
        smode = self.attr.get(SHARING_MODE, None)
        if (smode is not None and
                smode in bp.DedicatedSharingMode.ALL_VALUES):
                return True

    def _shared_procs_specified(self):
        """Determine if shared procs should be configured.

        General methodology is to try to check everything that would
        indicate shared processors first, then dedicated, and finally
        just default to shared if we can't determine either way.
        """
        if self.attr.get(DED_PROCS, None) is not None:
            return not DedicatedProc.convert_value(self.attr[DED_PROCS])

        # Check each key that would indicate sharing procs
        if self._shared_proc_keys_specified():
            return True

        # Check for dedicated sharing mode
        if self._dedicated_proc_keys_specified():
            return False

        # Default is to use shared if not proven otherwise
        return True

    def build(self):
        # Build a minimimal LPAR, the real work will be done in rebuild
        std = self.stdz.general()

        if std[ENV] == bp.LPARType.VIOS:
            lpar_w = vios.VIOS.bld(
                self.adapter, std[NAME],
                bp.PartitionMemoryConfiguration.bld(self.adapter, 0),
                bp.PartitionProcessorConfiguration.bld_dedicated(
                    self.adapter, 0),
                io_cfg=bp.PartitionIOConfiguration.bld(self.adapter, 0))
        else:
            lpar_w = lpar.LPAR.bld(
                self.adapter, std[NAME],
                bp.PartitionMemoryConfiguration.bld(self.adapter, 0),
                bp.PartitionProcessorConfiguration.bld_dedicated(
                    self.adapter, 0),
                io_cfg=bp.PartitionIOConfiguration.bld(self.adapter, 0),
                env=std[ENV])

        # Only set the uuid if one is sent in, otherwise it will be set
        # by PowerVM
        if std.get(UUID) is not None:
            lpar_w.uuid = std[UUID]

        if std.get(ID) is not None:
            lpar_w._id(std[ID])

        return self.rebuild(lpar_w)

    def rebuild(self, lpar_w):
        # Build the memory section
        mem_cfg = self.build_mem()

        # Build proc section
        # Determine if using shared or dedicated processors
        if self._shared_procs_specified():
            proc_cfg = self.build_shr_proc()
        else:
            proc_cfg = self.build_ded_proc()

        # Update any general attributes
        std = self.stdz.general()
        lpar_w.name = std[NAME]
        lpar_w.avail_priority = std[AVAIL_PRIORITY]
        lpar_w.proc_compat_mode = std[PROC_COMPAT]
        # Host may not be capable of SRR, so only add it if it's in the
        # standardized attributes
        if std.get(SRR_CAPABLE) is not None:
            lpar_w.srr_enabled = std[SRR_CAPABLE]
        io_cfg = bp.PartitionIOConfiguration.bld(self.adapter,
                                                 std[MAX_IO_SLOTS])

        # Now start replacing the sections
        lpar_w.mem_config = mem_cfg
        lpar_w.proc_config = proc_cfg
        lpar_w.io_config = io_cfg

        # Add IBMi values if needed
        if lpar_w.env == bp.LPARType.OS400:
            lpar_w.io_config.tagged_io = bp.TaggedIO.bld(
                self.adapter, load_src=std[LOAD_SRC], console=std[CONSOLE],
                alt_load_src=std[ALT_LOAD_SRC])
            if std.get(RESTRICTED_IO) is not None:
                lpar_w.restrictedio = std[RESTRICTED_IO]

        return lpar_w



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""HttpError, the EntryWrapper for HttpErrorResponse ('web' namespace)."""

from oslo_log import log as logging

import pypowervm.const as pc
import pypowervm.wrappers.entry_wrapper as ewrap

LOG = logging.getLogger(__name__)

_REASON_CODE = 'ReasonCode'
_MESSAGE = 'Message'
_HTTP_STATUS = 'HTTPStatus'

# Error codes that indicate the VIOS is busy
_VIOS_BUSY_ERR_CODES = ['HSCL3205', 'VIOS0014']


@ewrap.EntryWrapper.pvm_type('HttpErrorResponse', ns=pc.WEB_NS)
class HttpError(ewrap.EntryWrapper):

    @property
    def status(self):
        return self._get_val_int(_HTTP_STATUS)

    @property
    def reason_code(self):
        return self._get_val_str(_REASON_CODE)

    @property
    def message(self):
        return self._get_val_str(_MESSAGE)

    def is_vios_busy(self):
        try:
            msg = self.message
            if any(code in msg for code in _VIOS_BUSY_ERR_CODES):
                return True

            return self._legacy_message_check(msg)
        except Exception:
            return False

    def _legacy_message_check(self, msg):
        # This logic is...unfortunate.  We have to parse messages for strings
        # (instead of keys).  But we will only do that if it is marked an
        # internal error.
        if self.status != pc.HTTPStatus.INTERNAL_ERROR:
            return False

        # The old message met the following criteria
        if ('VIOS' in msg and
                'is busy processing some other request' in msg):
            return True

        # The new message format is the following
        if 'The system is currently too busy' in msg:
            return True

        # All others, assume not busy
        return False



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

import pypowervm.entities as ent
from pypowervm.tasks import memory
import pypowervm.tests.test_fixtures as fx
from pypowervm.wrappers import job


class TestMemory(testtools.TestCase):
    """Unit Tests for Memory tasks."""

    def setUp(self):
        super(TestMemory, self).setUp()
        entry = ent.Entry({}, ent.Element('Dummy', None), None)
        self.mock_job = job.Job(entry)
        self.adpt = self.useFixture(fx.AdapterFx()).adpt

    @mock.patch('pypowervm.wrappers.job.Job.wrap')
    @mock.patch('pypowervm.wrappers.job.Job.run_job')
    @mock.patch('pypowervm.wrappers.job.Job.create_job_parameter')
    @mock.patch('pypowervm.wrappers.job.Job.get_job_results_as_dict')
    def test_calculate_memory_overhead_on_host(self, mock_job_dict_res,
                                               mock_job_p,
                                               mock_run_job,
                                               mock_job_w):
        """Performs a simple set of calculate_memory_overhead_on_host tests."""

        def _reset_mocks():
            mock_job_w.reset_mock()
            mock_job_p.reset_mock()
            mock_run_job.reset_mock()
            mock_job_dict_res.reset_mock()

        def raise_exc_se():
            raise Exception

        mock_job_w.return_value = self.mock_job
        mock_host_uuid = '1234'
        args = ['ManagedSystem', mock_host_uuid]
        kwargs = {'suffix_type': 'do', 'suffix_parm': ('QueryReservedMemory'
                                                       'RequiredForPartition')}

        # test empty job results dictionary with defaults
        mock_job_dict_res.return_value = {'RequiredMemory': None,
                                          'CurrentAvailableSystemMemory': None}
        overhead, avail = (memory.
                           calculate_memory_overhead_on_host(self.adpt,
                                                             mock_host_uuid))
        self.adpt.read.assert_called_once_with(*args, **kwargs)
        self.assertEqual(1, mock_job_w.call_count)
        self.assertEqual(6, mock_job_p.call_count)
        self.assertEqual(1, mock_run_job.call_count)
        self.assertEqual(1, mock_job_dict_res.call_count)
        self.assertEqual(512, overhead)
        self.assertEqual(None, avail)
        _reset_mocks()

        # test with desired mem and non empty job results dict
        mock_job_dict_res.return_value = {'RequiredMemory': 1024,
                                          'CurrentAvailableSystemMemory':
                                          32768}
        reserved_mem_data = {'desired_mem': 768, 'num_virt_eth_adapters': 2}
        kwargs2 = {'reserved_mem_data': reserved_mem_data}
        overhead, avail = (memory.
                           calculate_memory_overhead_on_host(self.adpt,
                                                             mock_host_uuid,
                                                             **kwargs2))
        self.assertEqual(6, mock_job_p.call_count)
        self.assertEqual((1024-768), overhead)
        self.assertEqual(32768, avail)
        _reset_mocks()

        # test defaults when run_job fails
        mock_run_job.side_effect = raise_exc_se
        overhead, avail = (memory.
                           calculate_memory_overhead_on_host(self.adpt,
                                                             mock_host_uuid))
        mock_job_p.assert_any_call('LogicalPartitionEnvironment',
                                   'AIX/Linux')
        mock_job_p.assert_any_call('DesiredMemory', '512')
        mock_job_p.assert_any_call('MaximumMemory', '32768')
        mock_job_p.assert_any_call('NumberOfVirtualEthernetAdapter', '2')
        mock_job_p.assert_any_call('NumberOfVirtualSCSIAdapter', '1')
        mock_job_p.assert_any_call('NumberOfVirtualFibreChannelAdapter', '1')
        self.assertEqual(512, overhead)
        self.assertEqual(None, avail)
        self.assertEqual(0, mock_job_dict_res.call_count)
        _reset_mocks()

        # test reserved_mem_data values are created as job params
        reserved_mem_data = {'desired_mem': 2048,
                             'max_mem': 65536,
                             'lpar_env': 'OS400',
                             'num_virt_eth_adapters': 4,
                             'num_vscsi_adapters': 5,
                             'num_vfc_adapters': 6}
        kwargs3 = {'reserved_mem_data': reserved_mem_data}
        overhead, avail = (memory.
                           calculate_memory_overhead_on_host(self.adpt,
                                                             mock_host_uuid,
                                                             **kwargs3))
        mock_job_p.assert_any_call('LogicalPartitionEnvironment',
                                   'OS400')
        mock_job_p.assert_any_call('DesiredMemory', '2048')
        mock_job_p.assert_any_call('MaximumMemory', '65536')
        mock_job_p.assert_any_call('NumberOfVirtualEthernetAdapter', '4')
        mock_job_p.assert_any_call('NumberOfVirtualSCSIAdapter', '5')
        mock_job_p.assert_any_call('NumberOfVirtualFibreChannelAdapter', '6')
        self.assertEqual(512, overhead)
        self.assertEqual(None, avail)



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.util as u
import pypowervm.wrappers.iocard as card
import pypowervm.wrappers.managed_system as ms
import pypowervm.wrappers.network as net


class TestSRIOVAdapter(twrap.TestWrapper):

    file = 'sys_with_sriov.txt'
    wrapper_class_to_test = ms.System

    def setUp(self):
        super(TestSRIOVAdapter, self).setUp()
        self.sriovs = self.dwrap.asio_config.sriov_adapters

    def test_list(self):
        self.assertEqual(4, len(self.sriovs))
        for sriov in self.sriovs:
            self.assertIsInstance(sriov, card.SRIOVAdapter)

    def test_attrs(self):
        desc = 'PCIe2 4-port (10Gb FCoE & 1GbE) SR&RJ45 Adapter'

        self.assertEqual('553713696', self.sriovs[0].id)
        self.assertEqual(desc, self.sriovs[0].description)
        self.assertEqual('U78C7.001.RCH0004-P1-C8',
                         self.sriovs[0].phys_loc_code)

    def test_ids(self):
        """Test .id and .sriov_adap_id."""
        # AdapterID inherited from IOAdapter
        self.assertEqual('553713696', self.sriovs[0].id)
        self.assertEqual(1, self.sriovs[0].sriov_adap_id)

        self.assertEqual('553713680', self.sriovs[2].id)
        self.assertIsNone(self.sriovs[2].sriov_adap_id)

    def test_mode(self):
        self.assertEqual('Sriov', self.sriovs[0].mode)
        # Test setter
        self.sriovs[0].mode = 'unknown'
        self.assertEqual('unknown', self.sriovs[0].mode)

    def test_state(self):
        self.assertEqual('Running', self.sriovs[0].state)

    def test_physical_ports(self):
        adapter = self.sriovs[0]

        phyports = adapter.phys_ports
        self.assertEqual(4, len(phyports))

        # Get converged and ethernet physical ports each
        conv_port, eth_port = self.sriovs[0].phys_ports[:3:2]

        # Converged physical ports test
        self.assertEqual(self.sriovs[0], conv_port.sriov_adap)
        self.assertEqual(self.sriovs[0].sriov_adap_id, conv_port.sriov_adap_id)

        self.assertEqual(None, conv_port.label)
        conv_port.label = 'updatedlabel'
        self.assertEqual('updatedlabel', conv_port.label)

        self.assertEqual('U78C7.001.RCH0004-P1-C8-T1', conv_port.loc_code)

        self.assertEqual(0, conv_port.port_id)

        self.assertEqual(None, conv_port.sublabel)
        conv_port.sublabel = 'updatedsublabel'
        self.assertEqual('updatedsublabel', conv_port.sublabel)

        self.assertEqual(True, conv_port.link_status)

        self.assertEqual(20, conv_port.cfg_max_lps)
        conv_port.cfg_max_lps = 40
        self.assertEqual(40, conv_port.cfg_max_lps)

        self.assertEqual(2, conv_port.cfg_lps)

        self.assertEqual(0.02, conv_port.min_granularity)

        self.assertEqual(20, conv_port.supp_max_lps)

        self.assertEqual(0.02, conv_port.allocated_capacity)

        # Ethernet physical ports test
        self.assertEqual(self.sriovs[0], eth_port.sriov_adap)
        self.assertEqual(self.sriovs[0].sriov_adap_id, eth_port.sriov_adap_id)

        self.assertEqual(None, eth_port.label)
        eth_port.label = 'updatedlabel'
        self.assertEqual('updatedlabel', eth_port.label)

        self.assertEqual('U78C7.001.RCH0004-P1-C8-T3', eth_port.loc_code)

        self.assertEqual(2, eth_port.port_id)

        self.assertEqual(None, eth_port.sublabel)
        eth_port.sublabel = 'updatedsublabel'
        self.assertEqual('updatedsublabel', eth_port.sublabel)

        self.assertEqual(True, eth_port.link_status)

        self.assertEqual(4, eth_port.cfg_max_lps)
        eth_port.cfg_max_lps = 40
        self.assertEqual(40, eth_port.cfg_max_lps)

        self.assertEqual(0.02, eth_port.min_granularity)

        self.assertEqual(4, eth_port.supp_max_lps)

        self.assertEqual(0.02, eth_port.allocated_capacity)

        self.assertEqual(card.SRIOVSpeed.E1G, eth_port.curr_speed)

        self.assertEqual(card.SRIOVPPMTU.E1500, eth_port.mtu)
        eth_port.mtu = card.SRIOVPPMTU.E9000
        self.assertEqual(card.SRIOVPPMTU.E9000, eth_port.mtu)

        self.assertFalse(eth_port.flow_ctl)
        eth_port.flow_ctl = True
        self.assertTrue(eth_port.flow_ctl)

        self.assertEqual(net.VSwitchMode.VEB, eth_port.switch_mode)
        eth_port.switch_mode = net.VSwitchMode.VEPA
        self.assertEqual(net.VSwitchMode.VEPA, eth_port.switch_mode)

    def test_physical_ports_no_vivify(self):
        """Don't accidentally vivify [Converged]EthernetPhysicalPorts.

        See https://bugs.launchpad.net/pypowervm/+bug/1617050
        This test case has to prove that, when EthernetPhysicalPorts doesn't
        exist in the XML, asking for phys_ports doesn't create it.
        """
        # 2nd and 3rd SRIOV adapters have no pports
        adp = self.sriovs[2]
        self.assertNotIn('<EthernetPhysicalPorts ', adp.toxmlstring())
        self.assertNotIn('<ConvergedEthernetPhysicalPorts ', adp.toxmlstring())
        self.assertEqual([], adp.phys_ports)
        self.assertNotIn('<EthernetPhysicalPorts ', adp.toxmlstring())
        self.assertNotIn('<ConvergedEthernetPhysicalPorts ', adp.toxmlstring())


class TestLogicalPort(twrap.TestWrapper):

    file = 'sriov_lp_feed.txt'
    wrapper_class_to_test = card.SRIOVEthLPort

    def test_logical_ports(self):
        # Verify logical port getters
        lport = self.dwrap
        self.assertEqual(654327810, lport.lport_id)
        self.assertEqual(1, lport.sriov_adap_id)
        self.assertFalse(lport.is_promisc)
        self.assertEqual('PHB 4098', lport.dev_name)
        self.assertEqual(0.02, lport.cfg_capacity)
        self.assertEqual(2, lport.pport_id)
        self.assertEqual(0, lport.pvid)
        self.assertEqual(u.VLANList.ALL, lport.allowed_vlans)
        self.assertEqual('8ACA227C6E00', lport.mac)
        self.assertEqual(u.MACList.NONE, lport.allowed_macs)
        self.assertEqual('000000000000', lport.cur_mac)
        self.assertEqual('U78CB.001.WZS0485-P1-C5-T3-S2', lport.loc_code)
        self.assertEqual(card.VNICPortUsage.NOT_VNIC, lport.vnic_port_usage)

        # Verify logical port setters
        lport._sriov_adap_id(2)
        self.assertEqual(2, lport.sriov_adap_id)
        lport._is_promisc('true')
        self.assertTrue(lport.is_promisc)
        lport._pport_id(3)
        self.assertEqual(3, lport.pport_id)
        lport.allowed_vlans = [1, 2, 3]
        self.assertEqual([1, 2, 3], lport.allowed_vlans)
        lport.allowed_macs = u.MACList.ALL
        self.assertEqual(u.MACList.ALL, lport.allowed_macs)
        lport._cfg_capacity(0.0)
        self.assertEqual(0.0, lport.cfg_capacity)
        lport._cfg_capacity(1.0)
        self.assertEqual(1.0, lport.cfg_capacity)

        # Verify setter validation
        self.assertRaises(ValueError, lport._cfg_capacity, '5.0%')
        self.assertRaises(ValueError, lport._cfg_capacity, '2')
        self.assertRaises(ValueError, lport._cfg_capacity, float(2))
        self.assertRaises(ValueError, lport._cfg_capacity, '1.01')
        self.assertRaises(ValueError, lport._cfg_capacity, float(1.01))
        self.assertRaises(ValueError, lport._cfg_capacity, '-0.01')
        self.assertRaises(ValueError, lport._cfg_capacity, float(-0.01))
        self.assertRaises(ValueError, lport._cfg_capacity, 'garbage')
        self.assertRaises(ValueError, lport._cfg_capacity, '0.72%')

        # Verify bld method
        # With default kwargs
        lport = card.SRIOVEthLPort.bld(
            self.adpt, 5, 6)
        self.assertEqual(5, lport.sriov_adap_id)
        self.assertFalse(lport.is_promisc)
        self.assertIsNone(lport.cfg_capacity)
        self.assertEqual(6, lport.pport_id)
        self.assertIsNone(lport.pvid)
        self.assertEqual(u.MACList.ALL, lport.allowed_vlans)
        self.assertIsNone(lport.mac)
        self.assertEqual(u.MACList.ALL, lport.allowed_macs)
        # With explicit kwargs
        lport = card.SRIOVEthLPort.bld(
            self.adpt, 5, 6, pvid=2230, mac='12:ab:34:CD:56:ef',
            allowed_vlans=[1, 2, 3], allowed_macs=u.MACList.NONE,
            is_promisc=True, cfg_capacity=0.05)
        self.assertEqual(5, lport.sriov_adap_id)
        self.assertTrue(lport.is_promisc)
        self.assertEqual(0.05, lport.cfg_capacity)
        self.assertEqual(6, lport.pport_id)
        self.assertEqual(2230, lport.pvid)
        self.assertEqual([1, 2, 3], lport.allowed_vlans)
        self.assertEqual('12AB34CD56EF', lport.mac)
        self.assertEqual(u.MACList.NONE, lport.allowed_macs)


class TestVNIC(twrap.TestWrapper):
    file = 'vnic_feed.txt'
    wrapper_class_to_test = card.VNIC

    def test_vnic_props(self):
        self.assertEqual('U8286.42A.21C1B6V-V10-C3', self.dwrap.drc_name)
        self.assertEqual(10, self.dwrap.lpar_id)
        self.assertEqual(7, self.dwrap.slot)

    def test_vnic_and_backdev_bld(self):
        # Defaults in kwargs
        vnic = card.VNIC.bld(self.adpt)
        backdevs = vnic.back_devs
        self.assertEqual(self.adpt, vnic.adapter)
        self.assertIsNone(vnic.pvid)
        self.assertIsNotNone(backdevs)
        self.assertEqual(0, len(backdevs))
        # Fields without setters
        self.assertIsNone(vnic.drc_name)
        self.assertIsNone(vnic.lpar_id)
        self.assertIsNone(vnic.capacity)
        # Fields with setters not invoked (because not specified)
        self.assertIsNone(vnic.slot)
        self.assertTrue(vnic._use_next_avail_slot_id)
        self.assertFalse(vnic._get_val_bool(card._VNIC_USE_NEXT_AVAIL_SLOT))
        self.assertEqual(u.VLANList.ALL, vnic.allowed_vlans)
        self.assertIsNone(vnic.mac)
        self.assertEqual(u.MACList.ALL, vnic.allowed_macs)
        self.assertFalse(vnic.auto_pri_failover)
        vnic.auto_pri_failover = True
        self.assertTrue(vnic.auto_pri_failover)
        vnic.auto_pri_failover = False
        self.assertFalse(vnic.auto_pri_failover)

        # Values in kwargs

        def build_href(sch, uuid, **kwargs):
            self.assertEqual('VirtualIOServer', sch)
            self.assertEqual([], kwargs['xag'])
            return 'http://' + uuid
        self.adpt.build_href.side_effect = build_href

        backdevs = [card.VNICBackDev.bld(self.adpt, 'vios_uuid', 3, 4),
                    card.VNICBackDev.bld(self.adpt, 'vios_uuid2', 5, 6,
                                         capacity=0.3456789, failover_pri=50)]
        vnic = card.VNIC.bld(
            self.adpt, pvid=7, slot_num=8, allowed_vlans=[1, 2],
            mac_addr='m:a:c',
            allowed_macs=['AB:12:CD:34:EF:56', '12ab34cd56ef'],
            back_devs=backdevs)
        backdevs = vnic.back_devs
        self.assertEqual(self.adpt, vnic.adapter)
        self.assertIsNone(vnic.drc_name)
        self.assertIsNone(vnic.lpar_id)
        self.assertEqual(7, vnic.pvid)
        self.assertEqual(8, vnic.slot)
        self.assertFalse(vnic._use_next_avail_slot_id)
        self.assertFalse(vnic._get_val_bool(card._VNIC_USE_NEXT_AVAIL_SLOT))
        self.assertEqual([1, 2], vnic.allowed_vlans)
        self.assertEqual(['AB12CD34EF56', '12AB34CD56EF'], vnic.allowed_macs)
        self.assertEqual('MAC', vnic.mac)
        self.assertIsNotNone(backdevs)
        self.assertEqual(2, len(backdevs))
        bd1, bd2 = backdevs
        self.assertEqual(self.adpt, bd1.adapter)
        self.adpt.build_href.assert_any_call('VirtualIOServer', 'vios_uuid',
                                             xag=[])
        self.assertEqual('http://vios_uuid', bd1.vios_href)
        self.assertEqual(3, bd1.sriov_adap_id)
        self.assertEqual(4, bd1.pport_id)
        self.assertIsNone(bd1.capacity)
        self.assertEqual(self.adpt, bd2.adapter)
        self.adpt.build_href.assert_any_call('VirtualIOServer', 'vios_uuid2',
                                             xag=[])
        self.assertEqual('http://vios_uuid2', bd2.vios_href)
        self.assertEqual(5, bd2.sriov_adap_id)
        self.assertEqual(6, bd2.pport_id)
        self.assertEqual(0.3457, bd2.capacity)
        self.assertIsNone(bd1.failover_pri)
        self.assertEqual(bd2.failover_pri, 50)
        bd1.failover_pri = 42
        bd2.failover_pri = 60
        self.assertEqual(bd1.failover_pri, 42)
        self.assertEqual(bd2.failover_pri, 60)

    def test_details_props_inner(self):
        self._test_details_props(self.dwrap._details)

    def test_details_props_outer(self):
        self._test_details_props(self.dwrap)

    def _test_details_props(self, dets):
        self.assertEqual(0, dets.pvid)
        dets.pvid = 123
        self.assertEqual(123, dets.pvid)
        self.assertEqual(u.VLANList.ALL, dets.allowed_vlans)
        dets.allowed_vlans = [1, 2, 3]
        self.assertEqual([1, 2, 3], dets.allowed_vlans)
        self.assertEqual(0.02, dets.capacity)

        def bad_capacity_setter(val):
            dets.capacity = val
        self.assertRaises(AttributeError, bad_capacity_setter, '0.04')

        def bad_vlans_setter(val):
            dets.allowed_vlans = val
        self.assertRaises(ValueError, bad_vlans_setter, 'foo')
        self.assertRaises(ValueError, bad_vlans_setter, ['a', 'b', 'c'])
        self.assertEqual('AE7A25E59A07', dets.mac)
        self.assertEqual(u.MACList.ALL, dets.allowed_macs)
        dets.allowed_macs = ['AB:12:cd:34:EF:56', '12ab34CD56ef']
        self.assertEqual(['AB12CD34EF56', '12AB34CD56EF'],
                         dets.allowed_macs)

        def bad_macs_setter(val):
            dets.allowed_macs = val
        self.assertRaises(ValueError, bad_macs_setter, 'foo')

    def test_backdev_props(self):
        self.assertEqual(2, len(self.dwrap.back_devs))
        backdev = self.dwrap.back_devs[0]
        self.assertEqual(
            'https://9.1.2.3:12443/rest/api/uom/ManagedSystem/1cab7366-6b73-34'
            '2c-9f43-ddfeb9f8edd3/VirtualIOServer/3E3F9BFC-C4EE-439E-B70A-1D36'
            '9213ED83', backdev.vios_href)
        self.assertEqual(1, backdev.sriov_adap_id)
        self.assertEqual(0, backdev.pport_id)
        self.assertEqual(
            'https://9.1.2.3:12443/rest/api/uom/VirtualIOServer/3E3F9BFC-C4EE-'
            '439E-B70A-1D369213ED83/SRIOVEthernetLogicalPort/af2a8c95-58d1-349'
            '6-9af6-8cd562f0e839', backdev.lport_href)
        self.assertEqual(0.02, backdev.capacity)
        self.assertFalse(backdev.is_active)
        self.assertEqual(backdev.status, card.VNICBackDevStatus.LINK_DOWN)
        self.assertTrue(self.dwrap.back_devs[1].is_active)
        self.assertEqual(self.dwrap.back_devs[1].status,
                         card.VNICBackDevStatus.OPERATIONAL)

if __name__ == "__main__":
    unittest.main()



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

ibmpowervm_opts = [
    cfg.IntOpt('pypowervm_update_collision_retries',
               default=5,
               help='Number of retries if an update operation failed due to '
                    'collision'),
    cfg.IntOpt('pypowervm_job_request_timeout',
               default=1800,
               help='Default timeout in seconds for PowerVM Job requests.'),
]

CONF = cfg.CONF
CONF.register_opts(ibmpowervm_opts)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import math

from oslo_log import log as logging

import pypowervm.const as c
from pypowervm.i18n import _
import pypowervm.log as lgc
from pypowervm.wrappers import base_partition as pvm_bp
from pypowervm.wrappers import job
from pypowervm.wrappers import managed_system as pvm_ms

LOG = logging.getLogger(__name__)


@lgc.logcall
def calculate_memory_overhead_on_host(adapter, host_uuid,
                                      reserved_mem_data={},
                                      lmb_size=None, default=512):
    """Calculate host memory overhead.

    A certain amount of additional memory, such as memory for firmware,
    is required from the host during the creation of an instance or while
    changing an instance's memory specifications. This method queries the
    host to get the reserved PHYP memory needed for an LPAR. The
    calculation is based off of the instance's max memory requested,
    network and I/O adapter configurations, and the host's HPT ratio.
    The job response contains a value for total memory required to create
    or change the LPAR, which is desired memory plus reserved PHYP memory.

    :param adapter: pypowervm adapter
    :param host_uuid: the UUID of the host
    :param reserve_mem_data: (Optional) dictionary with values for job params
        {'desired_mem': int,
         'max_mem': int,
         'lpar_env': 'AIX/Linux' OR 'OS400',
         'num_virt_eth_adapters': int,
         'num_vscsi_adapters': int,
         'num_vfc_adapters': int}
    :param lmb_size: (Optional) logical memory block size
    :param default: (Optional) default value to use for required memory
                    overhead value if there was an error with the job
    :return overhead: reserved host memory
    :return avail_mem: available host memory
    """
    # If desired memory and maximum memory are not known, this query is
    # part of calculating host stats, and specific configurations of an
    # instance is not known. This will use the config option for a default
    # maximum memory.
    desired_mem = reserved_mem_data.get('desired_mem', 512)
    max_mem = reserved_mem_data.get('max_mem', 32768)
    # If lmb size is given, round max mem up to be a multiple
    # of lmb size. If max_mem is 0, max_mem will be set to lmb size.
    if lmb_size is not None:
        max_mem = int(math.ceil((max_mem or 1) / float(lmb_size)) *
                      int(lmb_size))
    lpar_env = reserved_mem_data.get('lpar_env', pvm_bp.LPARType.AIXLINUX)
    num_virt_eth_adapter = reserved_mem_data.get('num_virt_eth_adapters', 2)
    num_vscsi_adapter = reserved_mem_data.get('num_vscsi_adapters', 1)
    num_vfc_adapter = reserved_mem_data.get('num_vfc_adapters', 1)
    job_wrapper = job.Job.wrap(adapter.read(pvm_ms.System.schema_type,
                                            host_uuid,
                                            suffix_type=c.SUFFIX_TYPE_DO,
                                            suffix_parm=('QueryReservedMemory'
                                                         'RequiredFor'
                                                         'Partition')))

    # Create job parameters
    job_parms = [job_wrapper.create_job_parameter(
        'LogicalPartitionEnvironment', lpar_env)]
    job_parms.append(job_wrapper.create_job_parameter(
        'DesiredMemory', str(desired_mem)))
    job_parms.append(job_wrapper.create_job_parameter(
        'MaximumMemory', str(max_mem)))
    job_parms.append(job_wrapper.create_job_parameter(
        'NumberOfVirtualEthernetAdapter', str(num_virt_eth_adapter)))
    job_parms.append(job_wrapper.create_job_parameter(
        'NumberOfVirtualSCSIAdapter', str(num_vscsi_adapter)))
    job_parms.append(job_wrapper.create_job_parameter(
        'NumberOfVirtualFibreChannelAdapter', str(num_vfc_adapter)))

    try:
        job_wrapper.run_job(host_uuid, job_parms=job_parms,
                            timeout=120)
        results = job_wrapper.get_job_results_as_dict()
    except Exception as error:
        LOG.error(_("Error obtaining host memory overhead for host "
                    "with UUID '%(host)s': %(error)s.") %
                  {'host': host_uuid, 'error': error})
        LOG.debug("Defaulting required memory overhead for host with UUID "
                  "'%s' to %d MB" % (host_uuid, default))
        return default, None
    required_mem = results.get('RequiredMemory')
    avail_mem = results.get('CurrentAvailableSystemMemory')
    if required_mem is not None:
        overhead = int(required_mem) - desired_mem
    else:
        overhead = default
    if avail_mem is not None:
        avail_mem = int(avail_mem)
    return overhead, avail_mem



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Manage NetworkBridge, TrunkAdapter, LoadGroup, SEA, etc."""

import abc
import copy

import six

from oslo_concurrency import lockutils as lock

from pypowervm import const as c
from pypowervm import exceptions as pvm_exc
from pypowervm import util as pvm_util
from pypowervm.utils import retry as pvm_retry
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import network as pvm_net
from pypowervm.wrappers import virtual_io_server as pvm_vios

_MAX_VLANS_PER_VEA = 20
_ENSURE_VLAN_LOCK = 'ensure_vlans_nb'


def ensure_vlans_on_nb(adapter, host_uuid, nb_uuid, vlan_ids):
    """Will make sure that the VLANs are assigned to the Network Bridge.

    This method will reorder the arbitrary VLAN IDs as needed (those which are
    the PVID of the VEAs, but not the primary VEA).

    VLANs are always added to VEAs that are 'non-primary' (not the first VEA).
    However, if the VLAN is on the primary VEA then it is left on the system.
    The only 'untagged' VLAN that is allowed is the primary VEA's PVID.

    If the VLAN specified is on another Network Bridge's VEA (which happens
    to be on the same virtual switch):
     - An error will be thrown if it is on the primary VEA.
     - It will be removed off the Network Bridge if it is on the non-primary
       VEA.

    This method will not remove VLAN IDs from the network bridge that aren't
    part of the vlan_ids list.  Instead, each VLAN is simply added to the
    Network Bridge's VLAN list.

    :param adapter: The pypowervm Adapter.
    :param host_uuid: The Server's UUID
    :param nb_uuid: The Network Bridge UUID.
    :param vlan_ids: The list of VLANs to ensure are on the Network Bridge.
    """
    _get_bridger(adapter, host_uuid).ensure_vlans_on_nb(nb_uuid, vlan_ids)


def ensure_vlan_on_nb(adapter, host_uuid, nb_uuid, vlan_id):
    """Will make sure that the VLAN is assigned to the Network Bridge.

    This method will reorder the arbitrary VLAN IDs as needed (those which are
    the PVID of the VEAs, but not the primary VEA).

    VLANs are always added to VEAs that are 'non-primary' (not the first VEA).
    However, if the VLAN is on the primary VEA then it is left on the system.
    The only 'untagged' VLAN that is allowed is the primary VEA's PVID.

    If the VLAN specified is on another Network Bridge's VEA (which happens
    to be on the same virtual switch):
     - An error will be thrown if it is on the primary VEA.
     - It will be removed off the Network Bridge if it is on the non-primary
       VEA.

    :param adapter: The pypowervm Adapter.
    :param host_uuid: The Server's UUID
    :param nb_uuid: The Network Bridge UUID.
    :param vlan_id: The VLAN identifier to ensure is on the system.
    """
    ensure_vlans_on_nb(adapter, host_uuid, nb_uuid, [vlan_id])


def remove_vlan_from_nb(adapter, host_uuid, nb_uuid, vlan_id,
                        fail_if_pvid=False, existing_nbs=None):
    """Will remove the VLAN from a given Network Bridge.

    :param adapter: The pypowervm Adapter.
    :param host_uuid: The host system UUID.
    :param nb_uuid: The Network Bridge UUID.
    :param vlan_id: The VLAN identifier.
    :param fail_if_pvid: If set to true, will raise an exception if this is
                         the PVID on a Network Bridge.
    :param existing_nbs: Optional.  If set, should be the existing network
                         bridge wrappers.  If not provided, will gather from
                         the system directly.
    """
    _get_bridger(adapter, host_uuid).remove_vlan_from_nb(nb_uuid, vlan_id,
                                                         fail_if_pvid,
                                                         existing_nbs)


def _get_bridger(adapter, host_uuid):
    """Returns the appropriate bridger for the action."""
    if adapter.traits.vnet_aware:
        return NetworkBridgerVNET(adapter, host_uuid)
    else:
        return NetworkBridgerTA(adapter, host_uuid)


@six.add_metaclass(abc.ABCMeta)
class NetworkBridger(object):
    """Defines the high level flows for the VLAN provisioning.

    This class has the generic flows, subclasses extend this for the
    derivations of vnet_aware and direct VLAN application.
    """

    def __init__(self, adapter, host_uuid):
        """Creates the bridger.

        :param adapter: The pypowervm Adapter.
        :param host_uuid: The host systems's UUID
        """
        self.adapter = adapter
        self.host_uuid = host_uuid
        self._orphan_map = None

    @lock.synchronized(_ENSURE_VLAN_LOCK)
    def ensure_vlans_on_nb(self, nb_uuid, vlan_ids):
        """Will make sure that the VLANs are assigned to the Network Bridge.

        This method will reorder the arbitrary VLAN IDs as needed (those which
        are the PVID of the TrunkAdapter, but not the primary TrunkAdapter).

        VLANs are always added to TrunkAdapters that are 'non-primary' (not the
        first TrunkAdapter).  However, if the VLAN is on the primary
        TrunkAdapter then it is left on the system. The only 'untagged' VLAN
        that is allowed is the primary TrunkAdapter's PVID.

        If the VLAN specified is on another Network Bridge's TrunkAdapter
        (which happens to be on the same virtual switch):
         - An error will be thrown if it is on the primary TrunkAdapter.
         - It will be removed off the Network Bridge if it is on the
           non-primary TrunkAdapter.

        This method will not remove VLAN IDs from the network bridge that
        aren't part of the vlan_ids list.  Instead, each VLAN is simply added
        to the Network Bridge's VLAN list.

        :param nb_uuid: The Network Bridge UUID.
        :param vlan_ids: The list of VLANs to ensure are on the Network Bridge.
        """
        self._ensure_vlans_on_nb_synch(nb_uuid, vlan_ids)

    @pvm_retry.retry(tries=60, delay_func=pvm_retry.STEPPED_RANDOM_DELAY)
    def _ensure_vlans_on_nb_synch(self, nb_uuid, vlan_ids):
        # Ensure the VLANs are ints, not strings.
        vlan_ids = [int(x) for x in vlan_ids]

        # Get the updated feed of NetworkBridges
        nb_wraps = pvm_net.NetBridge.get(
            self.adapter, parent_type=pvm_ms.System,
            parent_uuid=self.host_uuid)

        # Find the appropriate Network Bridge
        req_nb = pvm_util.find_wrapper(nb_wraps, nb_uuid)

        # Call down to the ensure_vlan_on_nb method only for the additions.
        new_vlans = []
        peer_nbs = self._find_peer_nbs(nb_wraps, req_nb)
        all_nbs_on_vs = [req_nb]
        all_nbs_on_vs.extend(peer_nbs)

        # Need to evaluate the status of each VLAN.
        for vlan_id in vlan_ids:
            # No action required.  The VLAN is already part of the bridge.
            if req_nb.supports_vlan(vlan_id):
                continue

            # If its supported by a peer...
            for peer_nb in peer_nbs:
                if peer_nb.supports_vlan(vlan_id):
                    # Remove the VLAN.
                    self._remove_vlan_from_nb_synch(peer_nb.uuid, vlan_id,
                                                    fail_if_pvid=True,
                                                    existing_nbs=nb_wraps)
                    break

            # If it is an arbitrary VLAN ID on our network.  This should be
            # very rare.  But if it does happen, we should re-order the VLANs
            # and then retry this whole method.
            if self._is_arbitrary_vid(vlan_id, all_nbs_on_vs):
                # Find a new arbitrary VLAN ID, and re-assign the original
                # value to this new one.
                other_vlans = (
                    vlan_ids + self._get_orphan_vlans(req_nb.vswitch_id))
                new_a_vid = self._find_new_arbitrary_vid(
                    all_nbs_on_vs, others=other_vlans)
                self._reassign_arbitrary_vid(vlan_id, new_a_vid, req_nb)
                return self._ensure_vlans_on_nb_synch(nb_uuid, vlan_ids)

            # At this point, we've done all the easy checks.  Next up is to
            # detect if it is an orphan.
            self._validate_orphan_on_ensure(vlan_id, req_nb.vswitch_id)

            # Lastly, if we're here...it must be a completely new VLAN.
            new_vlans.append(vlan_id)

        # If there are no new VLANs, no need to continue.
        if len(new_vlans) == 0:
            return

        # At this point, all of the new VLANs that need to be added are in the
        # new_vlans list.  Now we need to put them on load groups.
        self._add_vlans_to_nb(req_nb, all_nbs_on_vs, new_vlans)

        # At this point, the network bridge should just need to be updated.
        # The Load Groups on the Network Bridge should be correct.
        req_nb.update()

    @lock.synchronized(_ENSURE_VLAN_LOCK)
    def remove_vlan_from_nb(self, nb_uuid, vlan_id, fail_if_pvid=False,
                            existing_nbs=None):
        """Will remove the VLAN from a given Network Bridge.

        :param nb_uuid: The Network Bridge UUID.
        :param vlan_id: The VLAN identifier.
        :param fail_if_pvid: If set to true, will raise an exception if this is
                             the PVID on a Network Bridge.
        :param existing_nbs: Optional.  If set, should be the existing network
                             bridge wrappers.  If not provided, will gather
                             from the system directly.
        """
        self._remove_vlan_from_nb_synch(nb_uuid, vlan_id, fail_if_pvid,
                                        existing_nbs)

    @pvm_retry.retry(tries=60, delay_func=pvm_retry.STEPPED_RANDOM_DELAY)
    def _remove_vlan_from_nb_synch(self, nb_uuid, vlan_id, fail_if_pvid=False,
                                   existing_nbs=None):
        # Ensure we're working with an integer
        vlan_id = int(vlan_id)

        if existing_nbs is not None:
            nb_wraps = existing_nbs
        else:
            # Get the updated feed of NetworkBridges
            nb_wraps = pvm_net.NetBridge.get(
                self.adapter, parent_type=pvm_ms.System,
                parent_uuid=self.host_uuid)

        # Find our Network Bridge
        req_nb = pvm_util.find_wrapper(nb_wraps, nb_uuid)

        # Determine if we're trying to remove an arbitrary PVID off of the
        # network bridge.  If so, we need to get a new, available arbitrary
        # PVID and swap that in.
        if vlan_id in req_nb.arbitrary_pvids:
            # Need to find the peers on this vSwitch.  Arbitrary PVIDs can
            # only be used once per vSwitch
            all_nbs_on_vs = self._find_peer_nbs(nb_wraps, req_nb,
                                                include_self=True)

            # Find a new arbitrary VLAN ID and swap it to a new, available
            # value.  Need to get the orphans so that we do not assign to an
            # existing orphan VLAN.
            other_vlans = [vlan_id] + self._get_orphan_vlans(req_nb.vswitch_id)
            new_a_vid = self._find_new_arbitrary_vid(all_nbs_on_vs,
                                                     others=other_vlans)
            self._reassign_arbitrary_vid(vlan_id, new_a_vid, req_nb)
            return

        # If the VLAN is not on the bridge, no action
        if not req_nb.supports_vlan(vlan_id):
            return

        # Fail if we're the PVID.
        if fail_if_pvid and req_nb.load_grps[0].pvid == vlan_id:
            raise pvm_exc.PvidOfNetworkBridgeError(vlan_id=vlan_id)

        # If this is on the first load group/trunk adapter, we leave it.
        if (req_nb.load_grps[0].pvid == vlan_id or
                vlan_id in req_nb.load_grps[0].tagged_vlans or
                len(req_nb.load_grps) == 1):
            return

        # Rip the VLAN out of the wrapper element.
        self._remove_vlan_from_nb(req_nb, vlan_id)

        # Now update the network bridge.
        req_nb.update()

    def _is_arbitrary_vid(self, vlan, all_nbs):
        """Returns if the VLAN is an arbitrary PVID on any passed in network.

        :param vlan: The VLAN to check.
        :param all_nbs: All of the network bridges on a given vSwitch.
        :return: The network bridge that this is an arbitrary VLAN on.
        """
        for nb in all_nbs:
            if vlan in nb.arbitrary_pvids:
                return nb
        return None

    def _find_new_arbitrary_vid(self, all_nbs, others=()):
        """Returns a new VLAN ID that can be used as an arbitrary VID.

        :param all_nbs: All of the impacted network bridges.  Should all be on
                        the same vSwitch.
        :param others: List of other VLANs that should not be used as an
                       arbitrary.
        :return: A new VLAN ID that is not in use by any network bridge on this
                 vSwitch.
        """
        all_vlans = []

        for i_nb in all_nbs:
            all_vlans.extend(i_nb.list_vlans(pvid=True, arbitrary=True))
        all_vlans.extend(others)

        # Start at 4094, and walk down to find one that isn't already used.
        # Stop right before VLAN 1 as that is special in the system.
        for i in range(4094, 1, -1):
            if i not in all_vlans:
                return i
        return None

    @staticmethod
    def _find_peer_nbs(nb_wraps, nb, include_self=False):
        """Finds all of the peer (same vSwitch) Network Bridges.

        :param nb_wraps: List of pypowervm NetBridge wrappers.
        :param nb: The NetBridge to find.
        :param include_self: (Optional, Default False) If set to true, will
                             include the nb in the response list.
        :return: List of Network Bridges on the same vSwitch as the seed.  Does
                 not include the nb element.
        """
        # Find the vswitch to search for.
        vs_search_id = nb.seas[0].primary_adpt.vswitch_id

        ret = []
        for nb_elem in nb_wraps:
            # Don't include self.
            if nb.uuid == nb_elem.uuid and not include_self:
                continue

            # See if the vswitches match
            other_vs_id = nb_elem.seas[0].primary_adpt.vswitch_id
            if other_vs_id == vs_search_id:
                ret.append(nb_elem)
        return ret

    def _validate_orphan_on_ensure(self, vlan, vswitch_id):
        """Will throw an error if there is collision with VLAN and vSwitch.

        An orphan VLAN is defined as a VLAN (on a specific vSwitch) that is
        part of a VIOS, but not attached to a Network Bridge (ex. Shared
        Ethernet Adapter).

        :param vlan: The VLAN to query for.
        :param vswitch_id: The virtual switch identifier.  This is the short
                           number (0-15).
        :raises: OrphanVLANFoundOnProvision
        """
        orphan_map = self._get_orphan_map()

        # If no oprhans on the vSwitch, then we're fine.
        if not orphan_map.get(vswitch_id):
            return

        # Walk through each element.
        for vios_name, devices in orphan_map[vswitch_id].items():
            for dev_name, vlans in devices.items():
                if vlan in vlans:
                    raise pvm_exc.OrphanVLANFoundOnProvision(
                        dev_name=dev_name, vlan_id=vlan, vios=vios_name)

    def _get_orphan_vlans(self, vswitch_id):
        """Returns the list of orphan VLANs for a given vSwitch.

        See _validate_orphan_on_ensure for a definition of an orphan VLAN.

        :param vswitch_id: The virtual switch identifier.  This is the short
                           number (0-15).
        :return: List of orphan VLANs for the given vSwitch.
        """
        orphan_map = self._get_orphan_map()

        # If no orphans on the vSwitch, then return an empty list
        if orphan_map.get(vswitch_id) is None:
            return []

        orphan_vlans = set()
        for devices in orphan_map[vswitch_id].values():
            for dev_key in devices:
                orphan_vlans.update(devices[dev_key])
        return list(orphan_vlans)

    def _get_orphan_map(self):
        """Returns the orphan map.  See _build_orphan_map for format."""
        if self._orphan_map is None:
            self._orphan_map = self._build_orphan_map()
        return self._orphan_map

    def _build_orphan_map(self):
        """Builds the map of orphan VLANs per vSwitch.

        Will set the orphan_map variable.  The result will be of the following
        format:

        { vswitch_id: {'vios_name': { 'dev_name': [vlan_id1, vlan_id2]} } }

        Note: vswitch_id and vlan_id are int type.

        This call should be used sparingly.  The map is only built if
        provisioning a new VLAN or removing one.  The calls that this makes are
        expensive, but necessary for correctness.  This is why they are lazy
        loaded, as many calls may not even need this map.

        Also note that only trunk adapters are considered as orphans.  While
        there certainly could be non-trunked adapters present as well, they
        will not conflict.  So while strange, it's functional.

        :return: The orphan map.
        """
        # Wipe out the existing map.
        orphan_map = {}

        # Loop through all the VIOSes.
        vios_wraps = pvm_vios.VIOS.get(self.adapter, parent_type=pvm_ms.System,
                                       parent_uuid=self.host_uuid,
                                       xag=[c.XAG.VIO_NET])

        for vios_w in vios_wraps:
            # List all of the trunk adapters that are not part of the SEAs
            orphan_trunks = []
            for trunk in vios_w.trunk_adapters:
                # If the trunk has the same device ID as any of the SEAs
                # children, then it is not an orphan.
                for sea in vios_w.seas:
                    if sea.contains_device(trunk.dev_name):
                        break
                else:
                    orphan_trunks.append(trunk)

            # At this point, we know all the orphans for this VIOS.  Add them
            # to the map.
            for orphan_trunk in orphan_trunks:
                vlans = [orphan_trunk.pvid] + orphan_trunk.tagged_vlans
                self._put_orphan_in_map(
                    orphan_map, vios_w, orphan_trunk.vswitch_id,
                    orphan_trunk.dev_name, vlans)

        return orphan_map

    def _put_orphan_in_map(self, orphan_map, vios_w, vswitch_id, dev_name,
                           vlan_ids):
        # Make sure the orphan map is initialized and ready.
        if vswitch_id not in orphan_map:
            orphan_map[vswitch_id] = {}
        if vios_w.name not in orphan_map[vswitch_id]:
            orphan_map[vswitch_id][vios_w.name] = {}

        # We can't just replace the device name.  The name may be 'Unknown',
        # so we just keep appending.
        vio_part = orphan_map[vswitch_id][vios_w.name]
        if dev_name not in vio_part:
            vio_part[dev_name] = []
        vio_part[dev_name].extend(vlan_ids)

    def _reassign_arbitrary_vid(self, old_vid, new_vid, impacted_nb):
        """Moves the arbitrary VLAN ID from one Load Group to another.

        Should perform the actual update to the API.

        :param old_vid: The original arbitrary VLAN ID.
        :param new_vid: The new arbitrary VLAN ID.
        :param impacted_nb: The network bridge that is impacted.
        """
        raise NotImplementedError()

    def _add_vlans_to_nb(self, req_nb, all_nbs_on_vs, new_vlans):
        """Adds the VLANs to the Network Bridge Wrapper.

        :param req_nb: The NetworkBridge to add the VLANs to.  After this
                       method is complete, this req_nb will have the
                       appropriate information to perform an update to the
                       API.
        :param all_nbs_on_vs: List of all the network bridges on the virtual
                              switch.
        :param new_vlans: List of the new VLANs to put on the network bridge.
        """
        raise NotImplementedError()

    def _remove_vlan_from_nb(self, req_nb, vlan_id):
        """Removes the VLAN from the Network Bridge wrapper.

        :param req_nb: The Network Bridge.  Upon return, the wrapper should
                       not support the VLAN.
        :param vlan_id: The VLAN ID to remove.
        """
        raise NotImplementedError()


class NetworkBridgerVNET(NetworkBridger):
    """The virtual network aware NetworkBridger."""

    def _add_vlans_to_nb(self, req_nb, all_nbs_on_vs, new_vlans):
        """Adds the VLANs to the Network Bridge Wrapper.

        :param req_nb: The NetworkBridge to add the VLANs to.  After this
                       method is complete, this req_nb will have the
                       appropriate information to perform an update to the
                       API.
        :param all_nbs_on_vs: List of all the network bridges on the virtual
                              switch.
        :param new_vlans: List of the new VLANs to put on the network bridge.
        """
        # At this point, all of the new VLANs that need to be added are in the
        # new_vlans list.  Now we need to put them on load groups.
        vswitch_w = pvm_net.VSwitch.search(
            self.adapter, parent_type=pvm_ms.System,
            parent_uuid=self.host_uuid, one_result=True,
            switch_id=req_nb.vswitch_id)
        vnets = pvm_net.VNet.get(self.adapter, parent_type=pvm_ms.System,
                                 parent_uuid=self.host_uuid)

        for vlan_id in new_vlans:
            ld_grp = self._find_available_ld_grp(req_nb)
            vid_vnet = self._find_or_create_vnet(vnets, vlan_id, vswitch_w,
                                                 tagged=True)
            if ld_grp is None:
                # No load group means they're all full.  Need to create a new
                # Load Group.
                #
                # First, create a new 'non-tagging' virtual network
                other_vlans = (new_vlans +
                               self._get_orphan_vlans(req_nb.vswitch_id))
                arb_vid = self._find_new_arbitrary_vid(all_nbs_on_vs,
                                                       others=other_vlans)
                arb_vnet = self._find_or_create_vnet(vnets, arb_vid, vswitch_w,
                                                     tagged=False)

                # Now create the new load group...
                vnet_uris = [arb_vnet.related_href, vid_vnet.related_href]
                ld_grp = pvm_net.LoadGroup.bld(self.adapter, arb_vid,
                                               vnet_uris)

                # Append to network bridge...
                req_nb.load_grps.append(ld_grp)
            else:
                # There was a Load Group.  Just need to append this vnet to it.
                ld_grp.vnet_uri_list.append(vid_vnet.related_href)

    def _remove_vlan_from_nb(self, req_nb, vlan_id):
        """Removes the VLAN from the Network Bridge wrapper.

        :param req_nb: The Network Bridge.  Upon return, the wrapper should
                       not support the VLAN.
        :param vlan_id: The VLAN ID to remove.
        """
        # Find the matching load group.  Since the 'supports_vlan' passed
        # before, this will always find a value.
        matching_lg = None
        for lg in req_nb.load_grps[1:]:
            if vlan_id in lg.tagged_vlans:
                matching_lg = lg
                break

        # A load balanced bridge requires at least two load groups.  We can't
        # remove a load group from the network bridge if it is load balanced,
        # but only has two load groups...  Make sure if it is load balanced
        # we wouldn't be deleting a required load group.
        can_remove_for_lb = (len(req_nb.load_grps) > 2 or
                             not req_nb.load_balance)

        if can_remove_for_lb and len(matching_lg.tagged_vlans) == 1:
            # Remove the load group
            req_nb.load_grps.remove(matching_lg)
        else:
            # Else just remove that virtual network.  In the case of load
            # balancing, you may end up with the second load group being
            # just a place holder.  But this is required by the system.
            vnet_uri = self._find_vnet_uri_from_lg(matching_lg, vlan_id)
            matching_lg.vnet_uri_list.remove(vnet_uri)

    def _reassign_arbitrary_vid(self, old_vid, new_vid, impacted_nb):
        """Moves the arbitrary VLAN ID from one Load Group to another.

        :param old_vid: The original arbitrary VLAN ID.
        :param new_vid: The new arbitrary VLAN ID.
        :param impacted_nb: The network bridge that is impacted.
        """
        # Find the Load Group that has this arbitrary VID
        impacted_lg = None
        for ld_grp in impacted_nb.load_grps:
            if ld_grp.pvid == old_vid:
                impacted_lg = ld_grp
                break

        # For the _find_or_create_vnet, we need to query all the virtual
        # networks
        vswitch_w = pvm_net.VSwitch.search(
            self.adapter, parent_type=pvm_ms.System,
            parent_uuid=self.host_uuid, one_result=True,
            switch_id=impacted_nb.vswitch_id)
        vnets = pvm_net.VNet.get(self.adapter, parent_type=pvm_ms.System,
                                 parent_uuid=self.host_uuid)

        # Read the old virtual network
        old_uri = self._find_vnet_uri_from_lg(impacted_lg, old_vid)

        # Need to create the new Virtual Network
        new_vnet = self._find_or_create_vnet(vnets, new_vid, vswitch_w,
                                             tagged=False)

        # Now we need to clone the load group
        uris = copy.copy(impacted_lg.vnet_uri_list)
        if old_uri is not None:
            uris.remove(old_uri)
        uris.insert(0, new_vnet.related_href)
        new_lg_w = pvm_net.LoadGroup.bld(self.adapter, new_vid, uris)

        impacted_nb.load_grps.remove(impacted_lg)

        # Need two updates.  One to remove the load group.
        impacted_nb = impacted_nb.update()

        # A second to add the new load group in
        impacted_nb.load_grps.append(new_lg_w)
        impacted_nb = impacted_nb.update()

        # Now that the old vid is detached from the load group, need to delete
        # the Virtual Network (because it was 'tagged' = False).
        if old_uri is not None:
            self.adapter.delete_by_href(old_uri)

    def _find_or_create_vnet(self, vnets, vlan, vswitch,
                             tagged=True):
        """Will find (or create) the VNet.

        If the VirtualNetwork already exists but has a different tag attribute,
        this method will delete the old virtual network, and then recreate with
        the specified tagged value.

        :param vnets: The virtual network wrappers on the system.
        :param vlan: The VLAN to find.
        :param vswitch: The vSwitch wrapper.
        :param tagged: True if tagged traffic will flow through this network.
        :return: The VNet wrapper for this element.
        """
        # Look through the list of vnets passed in
        for vnet in vnets:
            if vnet.vlan == vlan and vnet.vswitch_id == vswitch.switch_id:
                if tagged == vnet.tagged:
                    return vnet
                else:
                    # We found a matching vNet, but the tag was wrong.  Need to
                    # delete it.
                    self.adapter.delete_by_href(vnet.href)
                    break

        # Could not find one.  Time to create it.
        name = 'VLAN%(vid)s-%(vswitch)s' % {'vid': str(vlan),
                                            'vswitch': vswitch.name}
        vnet_elem = pvm_net.VNet.bld(
            self.adapter, name, vlan, vswitch.related_href, tagged)
        return vnet_elem.create(parent_type=pvm_ms.System,
                                parent_uuid=self.host_uuid)

    def _find_available_ld_grp(self, nb):
        """Will return the Load Group that can support a new VLAN.

        This will be the load group with the lowest number of virtual networks
        on it.

        :param nb: The NetBridge to search through.
        :return: The 'best' LoadGroup within the NetBridge that can support a
                 new VLAN.  If all are full, will return None.  Best is
                 determined by 'one with fewest Virtual Networks'.
        """
        # Never provision to the first load group.  We do this to keep
        # consistency with how projects have done in past.
        if len(nb.load_grps) == 1:
            return None

        # Find the load group with the fewest VLANs.
        avail_count = 0
        cur_lg = None
        ld_grps = nb.load_grps[1:]
        for ld_grp in ld_grps:
            # If this Load Group is full, skip to the next.
            if len(ld_grp.vnet_uri_list) >= _MAX_VLANS_PER_VEA:
                continue

            avail_count += 1

            # If the load group hasn't been set - OR - this load group has
            # less than the previously set, update which we'll return
            if (cur_lg is None or
                    len(ld_grp.vnet_uri_list) < len(cur_lg.vnet_uri_list)):
                cur_lg = ld_grp

        # If load balancing is turned on, we have some further inspection to
        # do.
        #
        # When load balancing is enabled, the goal is to have the VLANs spread
        # evenly across the 'additional Load Groups'.  So you create pairs of
        # Load Groups and add VLANs back and forth between them.
        if nb.load_balance and cur_lg is not None:
            # If there is only one Load Group available, but we have an 'odd'
            # amount of load groups.  That signals that we need to create a new
            # Load Group because we're unbalanced.  Returning None will flag
            # to create a new Load Group to put the virtual network (VLAN) on.
            #
            # Being unbalanced will naturally occur once a pair is full of
            # VLANs.  In that case, the cur_lg would have been None and the
            # code would have created a single new Load Group for that new
            # VLAN.  The code does not want to create the pair at that time
            # because an empty Load Group supporting nothing is a waste.  So
            # this code only gets used when the second VLAN is being added,
            # thus balancing the Load Groups again.
            if avail_count == 1 and len(ld_grps) % 2 == 1:
                return None

        return cur_lg

    def _find_vnet_uri_from_lg(self, lg, vlan):
        """Finds the Virtual Network for a VLAN within a LoadGroup.

        :param lg: The LoadGroup wrapper.
        :param vlan: The VLAN within the Load Group to look for.
        :return: The Virtual Network URI for the vlan.  If not found within the
                 Load Group, None will be returned.
        """
        for vnet_uri in lg.vnet_uri_list:
            vnet_net = pvm_net.VNet.get_by_href(self.adapter, vnet_uri)
            if vnet_net.vlan == vlan:
                return vnet_net.related_href
        return None


class NetworkBridgerTA(NetworkBridger):
    """The Trunk Adapter aware NetworkBridger."""

    def _reassign_arbitrary_vid(self, old_vid, new_vid, impacted_nb):
        """Moves the arbitrary VLAN ID from one Load Group to another.

        Should perform the actual update to the API.

        :param old_vid: The original arbitrary VLAN ID.
        :param new_vid: The new arbitrary VLAN ID.
        :param impacted_nb: The network bridge that is impacted.
        """
        # Find the Trunk Adapters that has this arbitrary VID
        impacted_tas = (None, None)
        for ta in impacted_nb.seas[0].addl_adpts:
            if ta.pvid == old_vid:
                impacted_tas = self._trunk_list(impacted_nb, ta)
                break

        # For each Trunk Adapter, change the VID to the new value.
        for ta in impacted_tas:
                ta.pvid = new_vid

        # Call the update
        impacted_nb = impacted_nb.update()

    def _add_vlans_to_nb(self, req_nb, all_nbs_on_vs, new_vlans):
        """Adds the VLANs to the Network Bridge Wrapper.

        :param req_nb: The NetworkBridge to add the VLANs to.  After this
                       method is complete, this req_nb will have the
                       appropriate information to perform an update to the
                       API.
        :param all_nbs_on_vs: List of all the network bridges on the virtual
                              switch.
        :param new_vlans: List of the new VLANs to put on the network bridge.
        """
        # At this point, all of the new VLANs that need to be added are in the
        # new_vlans list.  Now we need to put them on trunk adapters.
        vswitch_w = pvm_net.VSwitch.search(
            self.adapter, parent_type=pvm_ms.System,
            parent_uuid=self.host_uuid, one_result=True,
            switch_id=req_nb.vswitch_id)
        for vlan_id in new_vlans:
            trunks = self._find_available_trunks(req_nb)

            if trunks is None:
                # No trunk adapter list means they're all full.  Need to create
                # a new Trunk Adapter (or pair) for the new VLAN.
                other_vlans = (new_vlans +
                               self._get_orphan_vlans(req_nb.vswitch_id))
                arb_vid = self._find_new_arbitrary_vid(all_nbs_on_vs,
                                                       others=other_vlans)

                for sea in req_nb.seas:
                    trunk = pvm_net.TrunkAdapter.bld(
                        self.adapter, arb_vid, [vlan_id], vswitch_w,
                        trunk_pri=sea.primary_adpt.trunk_pri)
                    sea.addl_adpts.append(trunk)
            else:
                # Available trunks were found.  Add the VLAN to each
                for trunk in trunks:
                    trunk.tagged_vlans.append(vlan_id)

    def _remove_vlan_from_nb(self, req_nb, vlan_id):
        """Removes the VLAN from the Network Bridge wrapper.

        :param req_nb: The Network Bridge.  Upon return, the wrapper should
                       not support the VLAN.
        :param vlan_id: The VLAN ID to remove.
        """
        # Find the matching trunk adapter.
        matching_tas = None
        for trunk in req_nb.seas[0].addl_adpts:
            if vlan_id in trunk.tagged_vlans:
                matching_tas = self._trunk_list(req_nb, trunk)
                break

        # A load balanced SEA requires at least a primary adapter and at least
        # one additional adapter.  We can't remove a trunk from the SEA if it
        # is load balanced, but only has a single additional
        can_remove_for_lb = (len(req_nb.seas[0].addl_adpts) > 1
                             if req_nb.load_balance else True)

        for matching_ta in matching_tas:
            if len(matching_ta.tagged_vlans) == 1 and can_remove_for_lb:
                # Last VLAN, so it can be removed from the SEA.
                for sea in req_nb.seas:
                    if matching_ta in sea.addl_adpts:
                        sea.addl_adpts.remove(matching_ta)
                        break
            else:
                # Otherwise, we just remove it from the list.
                matching_ta.tagged_vlans.remove(vlan_id)

    def _find_peer_trunk(self, nb, ta):
        """Finds the peer adapter when the network bridge is failover ready.

        When a Network Bridge is set up for failover, there are two SEAs.  Each
        is essentially a mirror of each other, but are on different Virtual
        I/O Servers.  This means identical Trunk Adapters - but a different
        physical adapters.

        This method finds the 'peer' adapter, that happens to be on a different
        I/O Server.

        :param nb: The network bridge wrapper.
        :param ta: The Trunk Adapter from the first SEA in the network bridge.
        :return: The peer adapter per the above criteria.  If the network
                 bridge is not set up for failover, then None is returned.
        """
        if len(nb.seas) <= 1:
            return None

        sea = nb.seas[1]
        if sea.primary_adpt.pvid == ta.pvid:
            return sea.primary_adpt

        for addl_adpt in sea.addl_adpts:
            if addl_adpt.pvid == ta.pvid:
                return addl_adpt

        return None

    def _trunk_list(self, nb, ta):
        """For a given trunk adapter, builds the list of trunks to modify.

        :param nb: The network bridge wrapper.
        :param ta: The Trunk Adapter from the first SEA in the network bridge.
        :return: List of trunk adapters.  Includes the peer.  If no peer, then
                 only one element is returned in the list.
        """
        peer = self._find_peer_trunk(nb, ta)
        if peer:
            return [ta, peer]
        else:
            return [ta]

    def _find_available_trunks(self, nb):
        """Will return a list of Trunk Adapters that can support a new VLAN.

        Finds the set of trunk adapters with the lowest number of VLANs on
        it.

        :param nb: The NetBridge to search through.
        :return: A set of trunk adapters that can support the new VLAN.  A set
                 is returned as there may be multiple Virtual I/O Servers that
                 support it.  Each I/O Server may have a trunk to update.  If
                 No available Trunk Adapters are found, then None is returned.
        """
        # Find a trunk with the lowest amount of VLANs on it.
        cur_min = None
        avail_count = 0
        for trunk in nb.seas[0].addl_adpts:
            # If this trunk has maxed out its VLANs, skip to next.
            if len(trunk.tagged_vlans) >= _MAX_VLANS_PER_VEA:
                continue

            # This could definitely support it...
            avail_count += 1

            # But, is it the best?
            if (cur_min is None or
                    len(trunk.tagged_vlans) < len(cur_min.tagged_vlans)):
                cur_min = trunk

        # If load balancing is turned on, we have some further inspection to
        # do.
        #
        # When load balancing is enabled, the goal is to have the VLANs spread
        # evenly across the 'additional trunk adapters'.  So you create pair
        # and add VLANs back and forth between them.
        if nb.load_balance and cur_min is not None:
            # If there is only one set of trunk adapters available, but we have
            # an 'odd' amount of trunk adapters, that signals that we need to
            # create a new Trunk Adapter because we're unbalanced.  Returning
            # None will flag to create a new Trunk Adapter to put the VLAN on.
            #
            # Being unbalanced will naturally occur once a pair is full of
            # VLANs.  In that case, the cur_min would have been None and the
            # code would have created a single set of Trunk Adapters for that
            # new VLAN.  The code does not want to create the pair at that time
            # because an empty Trunk Adapter supporting nothing is a waste.  So
            # this code only gets used when the second VLAN is being added,
            # thus balancing the Trunk Adapters again.
            if avail_count == 1 and len(nb.seas[0].addl_adpts) % 2 == 1:
                return None

        # Return the trunk list if we have a trunk adapter, otherwise just
        # return None
        return self._trunk_list(nb, cur_min) if cur_min is not None else None
